from math import isclose

import pytest

from conops import Battery
from conops.common.enums import ChargeState

"""Unit tests for conops.battery.Battery"""


class TestBatteryValidation:
    def test_validator_sets_watthour_calculation(self, batt_20wh):
        b = batt_20wh
        assert b.watthour == 20

    def test_init_charge_level_equals_watthour(self, batt_20wh):
        b = batt_20wh
        assert isclose(b.charge_level, b.watthour)

    def test_battery_level_is_full_on_init(self, batt_20wh):
        b = batt_20wh
        assert b.battery_level == pytest.approx(1.0)

    def test_provided_watthour_is_preserved_value(self, batt_20wh):
        b = Battery(amphour=2, voltage=10, watthour=123.0)
        assert b.watthour == 123.0

    def test_provided_watthour_sets_initial_charge_level(self, batt_20wh):
        b = Battery(amphour=2, voltage=10, watthour=123.0)
        assert isclose(b.charge_level, 123.0)


class TestBatteryChargeAndDrain:
    def test_initial_charge_level_full(self, batt_1wh):
        b = batt_1wh
        assert isclose(b.charge_level, 1.0)

    def test_charge_increases_to_cap(self, batt_1wh):
        b = batt_1wh
        b.charge_level = 0.25
        b.charge(power=3600, period=1)  # Charge 1 Wh -> should cap at 1.0
        assert isclose(b.charge_level, 1.0)

    def test_charge_does_not_exceed_watthour_on_subsequent_charge(self, batt_1wh):
        b = batt_1wh
        b.charge_level = 0.25
        b.charge(power=3600, period=1)
        b.charge(power=7200, period=1)
        assert isclose(b.charge_level, 1.0)

    def test_drain_reduces_charge_level_to_zero(self, batt_1wh):
        b = batt_1wh
        b.drain(power=3600, period=1)
        assert isclose(b.charge_level, 0.0)

    def test_drain_return_false_when_empty(self, batt_1wh):
        b = batt_1wh
        b.drain(power=3600, period=1)
        assert b.drain(power=3600, period=1) is False

    def test_battery_level_fraction_when_setting_charge_level(self, batt_20wh):
        b = batt_20wh
        b.charge_level = 10.0
        assert isclose(b.battery_level, 0.5)

    def test_battery_level_fraction_decreases_after_drain(self, batt_20wh):
        b = batt_20wh
        b.charge_level = 10.0
        b.drain(power=3600, period=1)
        assert isclose(b.battery_level, 9.0 / 20.0)


class TestBatteryAlerts:
    def test_default_battery_level_full(self, default_battery):
        b = default_battery
        assert b.battery_level == pytest.approx(1.0)

    def test_default_battery_alert_is_false(self, default_battery):
        b = default_battery
        assert b.battery_alert is False

    def test_default_emergency_recharge_is_false(self, default_battery):
        b = default_battery
        assert b.emergency_recharge is False

    def test_battery_level_below_max_depth_of_discharge(self, default_battery):
        b = default_battery
        b.charge_level = b.watthour * 0.2  # 20% charge level (below 70% minimum)
        min_allowed_charge = 1.0 - b.max_depth_of_discharge  # 0.7
        assert b.battery_level < min_allowed_charge

    def test_battery_alert_true_when_below_max_depth_of_discharge(
        self, default_battery
    ):
        b = default_battery
        b.charge_level = b.watthour * 0.6
        assert b.battery_alert is True

    def test_emergency_recharge_true_when_below_max_depth_of_discharge(
        self, default_battery
    ):
        b = default_battery
        b.charge_level = b.watthour * 0.6
        # Trigger battery_alert to set emergency_recharge state
        _ = b.battery_alert
        assert b.emergency_recharge is True

    def test_battery_level_below_recharge_threshold(self, default_battery):
        b = default_battery
        b.charge_level = b.watthour * 0.94
        assert b.battery_level < b.recharge_threshold

    def test_battery_alert_true_when_below_recharge_threshold(self, default_battery):
        b = default_battery
        # First discharge below max_depth_of_discharge to trigger emergency_recharge
        b.charge_level = b.watthour * 0.6
        _ = b.battery_alert
        # Now recharge to test the recharge_threshold condition
        b.charge_level = b.watthour * 0.94
        assert b.battery_alert is True

    def test_emergency_recharge_true_when_below_recharge_threshold(
        self, default_battery
    ):
        b = default_battery
        # First discharge below max_depth_of_discharge to trigger emergency_recharge
        b.charge_level = b.watthour * 0.6
        _ = b.battery_alert
        # Now recharge to test the recharge_threshold condition
        b.charge_level = b.watthour * 0.94
        assert b.emergency_recharge is True

    def test_battery_alert_cleared_above_recharge_threshold(self, default_battery):
        b = default_battery
        b.charge_level = b.watthour * 0.96
        assert b.battery_alert is False

    def test_emergency_recharge_cleared_above_recharge_threshold(self, default_battery):
        b = default_battery
        b.charge_level = b.watthour * 0.96
        assert b.emergency_recharge is False


class TestBatteryChargeState:
    def test_initial_charge_state_not_charging(self, batt_20wh):
        b = batt_20wh
        assert b.charge_state == ChargeState.NOT_CHARGING

    def test_charge_state_charging_when_power_positive_and_not_full(self, batt_20wh):
        b = batt_20wh
        b.charge_level = 10.0  # Not full
        b.charge(power=100, period=1)
        assert b.charge_state == ChargeState.CHARGING

    def test_charge_state_trickle_when_full_and_charging(self, batt_20wh):
        b = batt_20wh
        # Already full, charge again
        b.charge(power=100, period=1)
        assert b.charge_state == ChargeState.TRICKLE

    def test_charge_state_not_charging_when_power_zero(self, batt_20wh):
        b = batt_20wh
        b.charge_level = 10.0
        b.charge(power=0, period=1)
        assert b.charge_state == ChargeState.NOT_CHARGING

    def test_drain_clamps_charge_level_to_zero_when_over_drained(self, batt_1wh):
        b = batt_1wh
        b.drain(power=3600, period=2)  # Drain 2 Wh from 1 Wh battery
        assert isclose(b.charge_level, 0.0)
        assert isclose(b.battery_level, 0.0)



from pydantic import BaseModel


class PowerDraw(BaseModel):
    """
    Power draw characteristics for a given subsystem, with option to specify
    different power draws based on operational modes.

    Supports eclipse-aware power consumption where different power levels can be
    specified for eclipse conditions (e.g., heaters drawing more power in eclipse).
    """

    nominal_power: float = 50  # Watts
    peak_power: float = 300
    power_mode: dict[int, float] = {}
    eclipse_power: float | None = None
    eclipse_power_mode: dict[int, float] = {}

    def power(self, mode: int | None = None, in_eclipse: bool = False) -> float:
        """Get power draw for the given mode and eclipse state.

        Args:
            mode: Operational mode (None for nominal)
            in_eclipse: Whether spacecraft is in eclipse

        Returns:
            Power draw in watts
        """
        # If in eclipse and eclipse power is configured, use eclipse values
        if in_eclipse:
            if mode is not None and mode in self.eclipse_power_mode:
                return self.eclipse_power_mode[mode]
            if self.eclipse_power is not None:
                return self.eclipse_power

        # Otherwise use normal power values
        if mode is None:
            return self.nominal_power
        return self.power_mode.get(mode, self.nominal_power)



"""Basic DITL timeline visualization with core spacecraft telemetry."""

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.font_manager import FontProperties

from ..config.visualization import VisualizationConfig

if TYPE_CHECKING:
    from ..ditl.ditl_mixin import DITLMixin


def plot_ditl_telemetry(
    ditl: "DITLMixin",
    figsize: tuple[float, float] = (10, 8),
    config: VisualizationConfig | None = None,
) -> tuple[Figure, list[Axes]]:
    """Plot basic DITL timeline with core spacecraft telemetry.

    Creates a 7-panel figure showing:
    - RA (Right Ascension)
    - Dec (Declination)
    - ACS Mode
    - Battery charge level with DoD limit
    - Solar panel illumination
    - Power consumption (with subsystem breakdown if available)
    - Observation ID

    Args:
        ditl: DITLMixin instance containing simulation telemetry data.
        figsize: Tuple of (width, height) for the figure size. Default: (10, 8)
        config: VisualizationConfig object. If None, uses ditl.config.visualization if available.

    Returns:
        tuple: (fig, axes) - The matplotlib figure and list of axes objects.

    Example:
        >>> from conops.ditl import QueueDITL
        >>> from conops.visualization import plot_ditl_telemetry
        >>> ditl = QueueDITL(config=config)
        >>> ditl.calc()
        >>> fig, axes = plot_ditl_telemetry(ditl)
        >>> plt.show()
    """
    # Resolve config: if the provided config is not a VisualizationConfig instance,
    # then try to use ditl.config.visualization if it's a VisualizationConfig, else use defaults.
    if not isinstance(config, VisualizationConfig):
        if (
            hasattr(ditl, "config")
            and hasattr(ditl.config, "visualization")
            and isinstance(ditl.config.visualization, VisualizationConfig)
        ):
            config = ditl.config.visualization
        else:
            config = VisualizationConfig()

    # Set default font settings
    font_family = config.font_family
    title_font_size = config.title_font_size
    label_font_size = config.label_font_size
    tick_font_size = config.tick_font_size
    title_prop = FontProperties(family=font_family, size=title_font_size, weight="bold")

    timehours = (np.array(ditl.utime) - ditl.utime[0]) / 3600

    fig = plt.figure(figsize=figsize)
    axes = []

    ax = plt.subplot(711)
    axes.append(ax)
    plt.plot(timehours, ditl.ra)
    ax.xaxis.set_visible(False)
    plt.ylabel("RA", fontsize=label_font_size, fontfamily=font_family)
    ax.set_title(
        f"Timeline for DITL Simulation: {ditl.config.name}", fontproperties=title_prop
    )

    ax = plt.subplot(712)
    axes.append(ax)
    ax.plot(timehours, ditl.dec)
    ax.xaxis.set_visible(False)
    plt.ylabel("Dec", fontsize=label_font_size, fontfamily=font_family)

    ax = plt.subplot(713)
    axes.append(ax)
    ax.plot(timehours, ditl.mode)
    ax.xaxis.set_visible(False)
    plt.ylabel("Mode", fontsize=label_font_size, fontfamily=font_family)

    ax = plt.subplot(714)
    axes.append(ax)
    ax.plot(timehours, ditl.batterylevel)
    ax.axhline(
        y=1.0 - ditl.config.battery.max_depth_of_discharge,
        color="r",
        linestyle="--",
    )
    ax.xaxis.set_visible(False)
    ax.set_ylim(0, 1)
    ax.set_ylabel("Batt. charge", fontsize=label_font_size, fontfamily=font_family)

    ax = plt.subplot(715)
    axes.append(ax)
    ax.plot(timehours, ditl.panel)
    ax.xaxis.set_visible(False)
    ax.set_ylim(0, 1)
    ax.set_ylabel("Panel Ill.", fontsize=label_font_size, fontfamily=font_family)

    ax = plt.subplot(716)
    axes.append(ax)
    # Check if subsystem power data is available
    if (
        hasattr(ditl, "power_bus")
        and hasattr(ditl, "power_payload")
        and ditl.power_bus
        and ditl.power_payload
    ):
        # Line plot showing power breakdown
        ax.plot(timehours, ditl.power_bus, label="Bus", alpha=0.8)
        ax.plot(timehours, ditl.power_payload, label="Payload", alpha=0.8)
        ax.plot(timehours, ditl.power, label="Total", linewidth=2, alpha=0.9)
        ax.legend(
            loc="upper right",
            fontsize=config.legend_font_size,
            prop={"family": font_family},
        )
    else:
        # Fall back to total power only
        ax.plot(timehours, ditl.power, label="Total")
    ax.set_ylim(0, max(ditl.power) * 1.1)
    ax.set_ylabel("Power (W)", fontsize=label_font_size, fontfamily=font_family)
    ax.xaxis.set_visible(False)

    ax = plt.subplot(717)
    axes.append(ax)
    ax.plot(timehours, ditl.obsid)
    ax.set_ylabel("ObsID", fontsize=label_font_size, fontfamily=font_family)
    ax.set_xlabel(
        "Time (hour of day)", fontsize=label_font_size, fontfamily=font_family
    )

    # Set tick font sizes for all axes
    for ax in axes:
        ax.tick_params(axis="both", which="major", labelsize=tick_font_size)
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontfamily(font_family)

    return fig, axes



"""Unit tests for Attitude Control System (ACS) class."""

from unittest.mock import Mock

import pytest

from conops import ACS, ACSMode


class DummyEphemeris:
    """Minimal mock ephemeris for testing."""

    def __init__(self):
        self.step_size = 1.0
        self.earth = [Mock(ra=Mock(deg=0.0), dec=Mock(deg=0.0))]

    def index(self, time):
        return 0


class TestACSInitialization:
    """Test ACS initialization and properties."""

    def test_acs_initialization(self, acs, mock_constraint):
        """Test that ACS initializes with correct defaults."""
        assert acs.constraint is mock_constraint
        assert acs.ra == 180.0  # Earth-opposite nadir pointing
        assert acs.dec == 0.0
        assert acs.roll == 0.0
        assert acs.obstype == "PPT"
        assert acs.last_slew is not None  # Initialized with boundary condition slew
        assert acs.last_ppt is None
        assert acs.acsmode == 0
        assert acs.current_pass is None
        assert acs.slew_dists == []

    def test_acs_requires_constraint(self, mock_config):
        """Test that ACS requires a constraint."""
        mock_config.constraint = None
        with pytest.raises(AssertionError, match="Constraint must be provided"):
            ACS(config=mock_config)

    def test_acs_requires_constraint_with_ephem(self, mock_config):
        """Test that ACS requires constraint with ephemeris."""
        mock_config.constraint.ephem = None
        with pytest.raises(AssertionError, match="Ephemeris must be set"):
            ACS(config=mock_config)


class TestACSAttributes:
    """Test ACS attribute access and manipulation."""

    def test_obstype_attribute(self, acs):
        """Test obstype attribute."""
        assert acs.obstype == "PPT"
        acs.obstype = "GSP"
        assert acs.obstype == "GSP"

    def test_constraint_attribute(self, acs, mock_constraint):
        """Test constraint attribute."""
        assert acs.constraint is mock_constraint

    def test_config_attribute(self, acs, mock_config):
        """Test config attribute."""
        assert acs.config is mock_config

    def test_solar_panel_attribute(self, acs):
        """Test solar_panel attribute exists."""
        assert acs.solar_panel is not None

    def test_ephem_attribute(self, acs):
        """Test ephem attribute exists."""
        assert acs.ephem is not None

    def test_last_slew_attribute(self, acs):
        """Test last_slew attribute."""
        assert acs.last_slew is not None  # Initialized with boundary condition slew

    def test_last_ppt_attribute(self, acs):
        """Test last_ppt attribute."""
        assert acs.last_ppt is None

    def test_current_pass_attribute(self, acs):
        """Test current_pass attribute."""
        assert acs.current_pass is None


class TestACSStateManagement:
    """Test ACS state management."""

    def test_acsmode_initial(self, acs):
        """Test default acsmode is SCIENCE."""
        assert acs.acsmode == ACSMode.SCIENCE

    def test_acsmode_set_slewing(self, acs):
        """Test setting acsmode to SLEWING."""
        acs.acsmode = ACSMode.SLEWING
        assert acs.acsmode == ACSMode.SLEWING

    def test_acsmode_set_saa(self, acs):
        """Test setting acsmode to SAA."""
        acs.acsmode = ACSMode.SAA
        assert acs.acsmode == ACSMode.SAA

    def test_acsmode_set_pass(self, acs):
        """Test setting acsmode to PASS."""
        acs.acsmode = ACSMode.PASS
        assert acs.acsmode == ACSMode.PASS

    def test_acsmode_set_charging(self, acs):
        """Test setting acsmode to CHARGING."""
        acs.acsmode = ACSMode.CHARGING
        assert acs.acsmode == ACSMode.CHARGING

    def test_acsmode_return_to_science(self, acs):
        """Test returning acsmode to SCIENCE."""
        acs.acsmode = ACSMode.SCIENCE
        assert acs.acsmode == ACSMode.SCIENCE

    def test_ra_dec_updates(self, acs):
        """Test that RA/Dec can be updated."""
        assert acs.ra == 180.0  # Earth-opposite nadir pointing
        assert acs.dec == 0.0
        acs.ra = 45.0
        acs.dec = 30.0
        assert acs.ra == 45.0
        assert acs.dec == 30.0

    def test_roll_updates(self, acs):
        """Test that roll can be updated."""
        assert acs.roll == 0.0
        acs.roll = 90.0
        assert acs.roll == 90.0

    def test_slew_dists_tracking(self, acs):
        """Test that slew_dists list is tracked."""
        assert acs.slew_dists == []
        acs.slew_dists.append(45.0)
        assert len(acs.slew_dists) == 1
        assert acs.slew_dists[0] == 45.0

    def test_rapid_state_changes(self, acs):
        """Test rapid state changes."""
        for _ in range(10):
            for mode in [
                ACSMode.SCIENCE,
                ACSMode.SLEWING,
                ACSMode.SAA,
                ACSMode.PASS,
                ACSMode.CHARGING,
            ]:
                acs.acsmode = mode
                assert acs.acsmode == mode


class TestACSPassRequestManagement:
    """Test pass request management."""

    def test_request_pass_basic(self, acs):
        """Test requesting a pass."""
        acs.passrequests.passes = []
        mock_pass = Mock()
        mock_pass.begin = 1000.0
        mock_pass.end = 2000.0
        mock_pass.obsid = 5

        acs.request_pass(mock_pass)
        assert mock_pass in acs.passrequests.passes

    def test_request_pass_multiple(self, acs):
        """Test requesting multiple non-overlapping passes."""
        acs.passrequests.passes = []
        mock_pass1 = Mock()
        mock_pass1.begin = 1000.0
        mock_pass1.end = 2000.0

        mock_pass2 = Mock()
        mock_pass2.begin = 3000.0
        mock_pass2.end = 4000.0

        acs.request_pass(mock_pass1)
        acs.request_pass(mock_pass2)
        assert len(acs.passrequests.passes) >= 2

    def test_request_pass_with_same_pass(self, acs):
        """Test that non-overlapping passes can be added to list."""
        acs.passrequests.passes = []
        passes_to_add = []
        for i in range(5):
            mock_pass = Mock()
            # Create non-overlapping passes: each starts after the previous one ends
            mock_pass.begin = 1000.0 + (i * 1500)  # Space passes 1500s apart
            mock_pass.end = 2000.0 + (i * 1500)  # Each pass is 1000s long
            passes_to_add.append(mock_pass)
            acs.request_pass(mock_pass)

        assert len(acs.passrequests.passes) == len(passes_to_add)


class TestACSStateTransitions:
    """Test state transitions during operations."""

    def test_full_state_cycle(self, acs):
        """Test cycling through all acsmode states."""
        for mode in [0, 1, 2, 3, 4]:
            acs.acsmode = mode
            assert acs.acsmode == mode



"""Test fixtures for target_queue subsystem tests."""

from datetime import datetime
from unittest.mock import Mock

import numpy as np
import pytest

from conops import Pointing, Queue


@pytest.fixture
def mock_target():
    """Fixture for a mock target."""
    target = Mock(spec=Pointing)
    target.merit = 100
    target.done = False
    target.ss_min = 60
    target.ss_max = 120
    target.slewtime = 10
    target.ra = 0
    target.dec = 0

    def reset_func():
        target.done = False

    target.reset = Mock(side_effect=reset_func)
    target.visible.return_value = True
    target.calc_slewtime = Mock(return_value=10)
    return target


@pytest.fixture
def mock_targets(mock_target):
    """Fixture for a list of mock targets."""
    targets = []
    for i in range(5):
        t = Mock(spec=Pointing)
        t.merit = 100 - i * 10
        t.done = False
        t.ss_min = 60
        t.ss_max = 120
        t.slewtime = 10
        t.ra = i * 10
        t.dec = i * 10

        # Create a closure to capture the target instance
        def create_reset_func(target_instance):
            def reset_func():
                target_instance.done = False

            return reset_func

        t.reset = Mock(side_effect=create_reset_func(t))
        t.visible.return_value = True
        t.calc_slewtime = Mock(return_value=10)
        targets.append(t)
    return targets


@pytest.fixture
def mock_config():
    """Fixture for a mock config."""
    config = Mock()
    config.constraint = Mock()
    config.spacecraft_bus = Mock()
    config.attitude_control = Mock()
    return config


@pytest.fixture
def queue_instance(mock_targets, mock_config):
    """Fixture for a Queue instance."""

    queue = Queue(config=mock_config)
    for target in mock_targets:
        queue.append(target)
    queue.ephem = Mock()
    # Mocking ephem timestamps for a day in the future
    queue.ephem.timestamp = np.array(
        [datetime.fromtimestamp(1762924800), datetime.fromtimestamp(1763011200)]
    )
    queue.ephem.datetimes = np.array(
        [datetime.fromtimestamp(1762924800), datetime.fromtimestamp(1763011200)]
    )

    queue.utime = 1762924800.0
    return queue



"""DITL Event logging data structure."""

from datetime import datetime, timezone

from pydantic import BaseModel, Field

from ..common import ACSMode


class DITLEvent(BaseModel):
    """
    A single event logged during a DITL simulation.

    Attributes:
        time: Unix timestamp of the event
        timestamp: ISO 8601 formatted timestamp string
        event_type: Category of event (e.g., 'PASS', 'SLEW', 'OBSERVATION', 'ERROR', 'INFO')
        description: Human-readable description of the event
        obsid: Observation ID associated with the event (if applicable)
        acs_mode: Current ACS mode at the time of the event (if applicable)
    """

    time: float = Field(..., description="Unix timestamp of the event")
    timestamp: str = Field(..., description="ISO 8601 formatted timestamp string")
    event_type: str = Field(
        ..., description="Category of event (e.g., 'PASS', 'SLEW', 'OBSERVATION')"
    )
    description: str = Field(..., description="Human-readable description of the event")
    obsid: int | None = Field(None, description="Observation ID (if applicable)")
    acs_mode: ACSMode | None = Field(
        None, description="Current ACS mode (if applicable)"
    )

    @classmethod
    def from_utime(
        cls,
        utime: float,
        event_type: str,
        description: str,
        obsid: int | None = None,
        acs_mode: ACSMode | None = None,
    ) -> "DITLEvent":
        """
        Create a DITLEvent from a unix timestamp.

        Args:
            utime: Unix timestamp
            event_type: Category of event
            description: Human-readable description
            obsid: Optional observation ID
            acs_mode: Optional ACS mode

        Returns:
            A new DITLEvent instance
        """
        dt = datetime.fromtimestamp(utime, tz=timezone.utc)
        timestamp = dt.isoformat(timespec="seconds")
        return cls(
            time=utime,
            timestamp=timestamp,
            event_type=event_type,
            description=description,
            obsid=obsid,
            acs_mode=acs_mode,
        )

    def __str__(self) -> str:
        """Format the event for display."""
        return f"[{self.timestamp}] {self.description}"



import numpy as np

# Constants
DTOR = np.pi / 180.0
DAY_SECONDS = 86_400

EARTH_OCCULT = 20
SUN_OCCULT = 50
ANTISUN_OCCULT = 180
MOON_OCCULT = 20

# Figure out the
PANEL_CONSTRAINT = 45  # Above 90 degrees means panel is not illuminated



"""Data management visualization utilities for CONOPS simulations."""

from datetime import datetime

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.font_manager import FontProperties

from conops.ditl.ditl import DITL
from conops.ditl.queue_ditl import QueueDITL

from ..config.visualization import VisualizationConfig


def plot_data_management_telemetry(
    ditl: QueueDITL | DITL,
    figsize: tuple[float, float] = (12, 10),
    show_summary: bool = True,
    config: VisualizationConfig | None = None,
) -> tuple[Figure, Axes]:
    """Plot comprehensive data management telemetry from a DITL simulation.

    Creates a multi-panel figure showing:
    1. Onboard recorder data volume over time
    2. Recorder fill fraction with alert thresholds
    3. Cumulative data generated
    4. Cumulative data downlinked
    5. Recorder alert timeline

    Args:
        ditl: DITL simulation object with data management telemetry.
        figsize: Tuple of (width, height) for the figure size. Default: (12, 10)
        show_summary: Whether to print summary statistics. Default: True
        config: VisualizationConfig object. If None, uses ditl.config.visualization if available.

    Returns:
        tuple: (fig, axes) - The matplotlib figure and axes objects.

    Example:
        >>> from conops.ditl import QueueDITL
        >>> from conops.visualization import plot_data_management_telemetry
        >>> ditl = QueueDITL(config=config)
        >>> ditl.calc()
        >>> fig, axes = plot_data_management_telemetry(ditl)
        >>> plt.show()
    """
    # Resolve config: use ditl.config.visualization if it's a real VisualizationConfig;
    # otherwise fall back to default VisualizationConfig.
    if not isinstance(config, VisualizationConfig):
        if (
            hasattr(ditl, "config")
            and hasattr(ditl.config, "visualization")
            and isinstance(ditl.config.visualization, VisualizationConfig)
        ):
            config = ditl.config.visualization
        else:
            config = VisualizationConfig()

    # Set default font settings
    font_family = config.font_family
    title_font_size = config.title_font_size
    # Create a FontProperties object for titles to ensure font family and size are applied
    title_prop = FontProperties(family=font_family, size=title_font_size, weight="bold")
    label_font_size = config.label_font_size
    tick_font_size = config.tick_font_size

    # Create figure with 5 subplots
    fig, axes = plt.subplots(5, 1, figsize=figsize)

    # Convert unix timestamps to datetime objects
    times = [datetime.fromtimestamp(t) for t in ditl.utime]

    # Plot 1: Recorder Volume (Gb)
    axes[0].plot(times, ditl.recorder_volume_gb, "b-", linewidth=2)
    axes[0].set_ylabel("Volume (Gb)", fontsize=label_font_size, fontfamily=font_family)
    axes[0].set_title("Onboard Recorder Data Volume", fontproperties=title_prop)
    axes[0].grid(True, alpha=0.3)
    axes[0].axhline(
        y=ditl.config.recorder.capacity_gb,
        color="r",
        linestyle="--",
        label=f"Capacity ({ditl.config.recorder.capacity_gb} Gb)",
    )
    axes[0].legend(prop={"family": font_family, "size": config.legend_font_size})

    # Plot 2: Recorder Fill Fraction
    axes[1].plot(times, ditl.recorder_fill_fraction, "g-", linewidth=2)
    axes[1].axhline(
        y=ditl.config.recorder.yellow_threshold,
        color="orange",
        linestyle="--",
        label=f"Yellow Threshold ({ditl.config.recorder.yellow_threshold:.0%})",
    )
    axes[1].axhline(
        y=ditl.config.recorder.red_threshold,
        color="red",
        linestyle="--",
        label=f"Red Threshold ({ditl.config.recorder.red_threshold:.0%})",
    )
    axes[1].set_ylabel(
        "Fill Fraction", fontsize=label_font_size, fontfamily=font_family
    )
    axes[1].set_title("Recorder Fill Level", fontproperties=title_prop)
    axes[1].grid(True, alpha=0.3)
    axes[1].legend(prop={"family": font_family, "size": config.legend_font_size})

    # Plot 3: Cumulative Data Generated (Gb)
    axes[2].plot(times, ditl.data_generated_gb, "m-", linewidth=2)
    axes[2].set_ylabel(
        "Data Generated (Gb)", fontsize=label_font_size, fontfamily=font_family
    )
    axes[2].set_title("Cumulative Data Generated", fontproperties=title_prop)
    axes[2].grid(True, alpha=0.3)

    # Plot 4: Cumulative Data Downlinked (Gb)
    axes[3].plot(times, ditl.data_downlinked_gb, "c-", linewidth=2)
    axes[3].set_ylabel(
        "Data Downlinked (Gb)", fontsize=label_font_size, fontfamily=font_family
    )
    axes[3].set_title("Cumulative Data Downlinked", fontproperties=title_prop)
    axes[3].grid(True, alpha=0.3)

    # Plot 5: Recorder Alert Timeline
    # Map integer alert levels to colors
    alert_colors = {0: "green", 1: "yellow", 2: "red"}
    colors = [alert_colors[alert] for alert in ditl.recorder_alert]
    axes[4].scatter(times, ditl.recorder_alert, c=colors, s=10, alpha=0.6)
    axes[4].set_yticks([0, 1, 2])
    axes[4].set_yticklabels(["None", "Yellow", "Red"])
    axes[4].set_ylabel("Alert Level", fontsize=label_font_size, fontfamily=font_family)
    axes[4].set_xlabel("Time", fontsize=label_font_size, fontfamily=font_family)
    axes[4].set_title("Recorder Alert Timeline", fontproperties=title_prop)
    axes[4].grid(True, alpha=0.3)

    # Set tick font sizes
    for ax in axes:
        ax.tick_params(axis="both", which="major", labelsize=tick_font_size)
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontfamily(font_family)

    plt.tight_layout()

    # Print summary statistics if requested
    if show_summary:
        _print_data_management_summary(ditl)

    return fig, axes


def _print_data_management_summary(ditl: QueueDITL | DITL) -> None:
    """Print summary statistics for data management.

    Args:
        ditl: DITL simulation object with data management telemetry.
    """
    total_generated = ditl.data_generated_gb[-1] if ditl.data_generated_gb else 0
    total_downlinked = ditl.data_downlinked_gb[-1] if ditl.data_downlinked_gb else 0
    final_volume = ditl.recorder_volume_gb[-1] if ditl.recorder_volume_gb else 0
    max_fill = max(ditl.recorder_fill_fraction) if ditl.recorder_fill_fraction else 0

    print("\n" + "=" * 70)
    print("DATA MANAGEMENT SUMMARY")
    print("=" * 70)
    print(f"Total Data Generated:    {total_generated:.2f} Gb")
    print(f"Total Data Downlinked:   {total_downlinked:.2f} Gb")
    print(f"Final Recorder Volume:   {final_volume:.2f} Gb")
    print(f"Max Fill Fraction:       {max_fill:.1%}")
    print(f"Recorder Capacity:       {ditl.config.recorder.capacity_gb:.2f} Gb")

    if total_generated > 0:
        downlink_efficiency = (total_downlinked / total_generated) * 100
        print(f"Downlink Efficiency:     {downlink_efficiency:.1f}%")

    print("=" * 70)



from .acs import ACS
from .acs_command import ACSCommand
from .emergency_charging import EmergencyCharging
from .passes import Pass, PassTimes
from .roll import optimum_roll, optimum_roll_sidemount
from .saa import SAA
from .slew import Slew

__all__ = [
    "ACS",
    "ACSCommand",
    "EmergencyCharging",
    "optimum_roll",
    "optimum_roll_sidemount",
    "Pass",
    "PassTimes",
    "SAA",
    "Slew",
]



"""Test fixtures for queue subsystem tests."""

from datetime import datetime, timedelta, timezone
from unittest.mock import Mock, patch

import numpy as np
import pytest
from astropy.time import Time  # type: ignore[import-untyped]

from conops import DAY_SECONDS, DumbQueueScheduler, QueueDITL
from conops.targets.plan import Plan


class DummyEphemeris:
    """Minimal mock ephemeris for testing."""

    def __init__(self):
        self.step_size = 3600.0  # Use 1 hour steps for faster tests
        # Cover 2018 day 331 (Nov 27) for 1 day
        base_time = datetime(2018, 11, 27, tzinfo=timezone.utc)
        # timestamp must be a list of datetime objects for helper functions
        # Use 1 hour steps: 24 steps for 1 day
        self.timestamp = [
            base_time + timedelta(seconds=i * 3600)
            for i in range(25)  # 24 hours + 1 for end time
        ]
        # Add earth and sun attributes for ACS initialization
        self.earth = [Mock(ra=Mock(deg=0.0), dec=Mock(deg=0.0)) for _ in range(25)]
        self.sun = [Mock(ra=Mock(deg=45.0), dec=Mock(deg=23.5)) for _ in range(25)]

    def index(self, time):
        """Mock index method."""
        return 0

    def in_eclipse(self, utime: float) -> bool:
        """Mock eclipse check - always return False (in sunlight)."""
        return False


@pytest.fixture
def mock_ephem():
    """Create a mock ephemeris object."""
    return DummyEphemeris()


@pytest.fixture
def mock_config():
    """Create a mock config with all required subsystems."""
    config = Mock()

    # Mock constraint
    config.constraint = Mock()
    config.constraint.ephem = DummyEphemeris()
    config.constraint.panel_constraint = Mock()
    config.constraint.panel_constraint.solar_panel = Mock()
    config.constraint.in_constraint = Mock(return_value=False)

    # Mock battery
    config.battery = Mock()
    config.battery.battery_level = 0.8
    config.battery.battery_alert = False
    config.battery.drain = Mock()
    config.battery.charge = Mock()

    # Mock spacecraft bus
    config.spacecraft_bus = Mock()
    config.spacecraft_bus.power = Mock(return_value=50.0)
    config.spacecraft_bus.attitude_control = Mock()
    config.spacecraft_bus.attitude_control.predict_slew = Mock(
        return_value=(10.0, [])
    )  # Return (distance, path)
    config.spacecraft_bus.attitude_control.slew_time = Mock(
        return_value=100.0
    )  # Return slew time in seconds

    # Mock payload
    config.payload = Mock()
    config.payload.power = Mock(return_value=30.0)

    # Mock solar panel
    config.solar_panel = Mock()
    config.solar_panel.optimal_charging_pointing = Mock(return_value=(45.0, 23.5))
    config.solar_panel.illumination_and_power = Mock(return_value=(0.5, 100.0))

    # Mock ground stations
    config.ground_stations = Mock()

    return config


@pytest.fixture
def queue_ditl(mock_config, mock_ephem):
    """Create a QueueDITL instance with mocked dependencies."""
    with (
        patch("conops.Queue") as mock_queue_class,
        patch("conops.PassTimes") as mock_passtimes,
        patch("conops.ACS") as mock_acs_class,
    ):
        # Mock PassTimes
        mock_pt = Mock()
        mock_pt.passes = []
        mock_pt.get = Mock()
        mock_pt.check_pass_timing = Mock(
            return_value={"start_pass": None, "end_pass": False, "updated_pass": None}
        )
        mock_passtimes.return_value = mock_pt

        # Mock ACS
        mock_acs = Mock()
        mock_acs.ephem = mock_ephem
        mock_acs.slewing = False
        mock_acs.inpass = False
        mock_acs.saa = None
        mock_acs.pointing = Mock(return_value=(0.0, 0.0, 0.0, 0))
        mock_acs.enqueue_command = Mock()
        mock_acs.passrequests = mock_pt
        mock_acs.slew_dists = []
        mock_acs.last_slew = None
        # Set acsmode to a real ACSMode enum value for logging
        from conops import ACSMode

        mock_acs.acsmode = ACSMode.SCIENCE
        # Mock the helper methods used in _fetch_new_ppt
        mock_target_request = Mock()
        mock_target_request.next_vis = Mock(return_value=1000.0)
        mock_acs._create_target_request = Mock(return_value=mock_target_request)
        mock_acs._initialize_slew_positions = Mock(return_value=True)
        mock_acs._is_slew_valid = Mock(return_value=True)
        mock_acs._calculate_slew_timing = Mock(return_value=1000.0)
        mock_acs_class.return_value = mock_acs

        # Mock solar panel in config (not ACS)
        mock_config.solar_panel.illumination_and_power = Mock(return_value=(0.5, 100.0))

        # Mock Queue
        mock_queue = Mock()
        mock_queue.get = Mock(return_value=None)
        mock_queue_class.return_value = mock_queue

        ditl = QueueDITL(config=mock_config, ephem=mock_ephem, queue=mock_queue)
        ditl.acs = mock_acs

        return ditl


@pytest.fixture
def mock_ephemeris():
    """Create a mock ephemeris object."""
    ephem = Mock()
    # 24 hours of data starting 2021-01-04
    start_time = Time("2021-01-04 00:00:00", scale="utc").unix
    timestamps = np.arange(start_time, start_time + DAY_SECONDS, 60)
    ephem.timestamp = Mock()
    ephem.timestamp.unix = timestamps
    return ephem


class MockPointing:
    """Mock Pointing class for testing."""

    def __init__(
        self,
        targetid=1,
        ra=45.0,
        dec=30.0,
        merit=100.0,
        ss_min=300,
        ss_max=600,
        name="",
    ):
        self.targetid = targetid
        self.ra = ra
        self.dec = dec
        self.merit = merit
        self.ss_min = ss_min  # Minimum exposure time
        self.ss_max = ss_max  # Maximum exposure time
        self.name = name or f"Target_{targetid}"
        self.done = False
        self.roll = 0.0
        self.slewtime = 0
        self.begin = 0
        self.end = 0

    def calc_slewtime(self, ra_from, dec_from):
        """Calculate slew time from prior position."""
        dist = np.sqrt((self.ra - ra_from) ** 2 + (self.dec - dec_from) ** 2)
        self.slewtime = max(0, int(dist / 0.5))  # Slew rate of 0.5 deg/sec

    def visible(self, start_time, end_time):
        """Check if target is visible during time window."""
        # Mock: always visible unless explicitly marked invisible
        return getattr(self, "_visible", True)


@pytest.fixture
def mock_queue(mock_ephemeris):
    """Create a mock queue with sample targets."""
    queue = Mock()
    queue.ephem = mock_ephemeris
    queue.targets = []

    # Add sample targets
    target1 = MockPointing(targetid=1, ra=45.0, dec=30.0, merit=100, ss_min=300)
    target2 = MockPointing(targetid=2, ra=90.0, dec=-45.0, merit=90, ss_min=300)
    target3 = MockPointing(targetid=3, ra=180.0, dec=60.0, merit=80, ss_min=300)

    queue.targets = [target1, target2, target3]
    queue.__len__ = Mock(return_value=len(queue.targets))
    queue.__getitem__ = Mock(side_effect=lambda i: queue.targets[i])

    def mock_get(ra, dec, utime):
        """Mock get method that returns next available target."""
        for target in queue.targets:
            if not target.done and target.merit > 0:
                target.calc_slewtime(ra, dec)
                if target.visible(utime, utime + target.slewtime + target.ss_max):
                    target.begin = int(utime)
                    target.end = int(utime + target.slewtime + target.ss_max)
                    return target
        return None

    def mock_meritsort(ra, dec):
        """Mock meritsort to sort by merit."""
        queue.targets.sort(key=lambda x: x.merit, reverse=True)

    queue.get = mock_get
    queue.meritsort = mock_meritsort
    queue.reset = Mock()

    return queue


@pytest.fixture
def scheduler(mock_queue, mock_ephemeris):
    """Create a DumbQueueScheduler instance."""
    begin = datetime(2021, 1, 4, tzinfo=timezone.utc)
    end = begin + timedelta(days=1)
    scheduler = DumbQueueScheduler(queue=mock_queue, begin=begin, end=end)
    scheduler.queue.ephem = mock_ephemeris
    # Override get to return None for basic tests
    scheduler.queue.get = Mock(return_value=None)
    return scheduler


@pytest.fixture
def mock_pointing():
    """Fixture to create a mock pointing object."""

    def _mock_pointing(targetid, ra, dec, merit, ss_min=None):
        pointing = Mock()
        pointing.targetid = targetid
        pointing.ra = ra
        pointing.dec = dec
        pointing.merit = merit
        if ss_min is not None:
            pointing.ss_min = ss_min
        pointing.done = False
        return pointing

    return _mock_pointing


@pytest.fixture
def make_target(mock_pointing):
    """Return a factory to create targets quickly."""

    def _make(targetid=1, ra=45.0, dec=30.0, merit=100, ss_min=300):
        t = mock_pointing(targetid=targetid, ra=ra, dec=dec, merit=merit, ss_min=ss_min)
        t.done = False
        return t

    return _make


@pytest.fixture
def make_targets(make_target):
    """Return a factory to create a list of targets."""

    def _make_many(count=3, start_ra=45.0):
        targets = []
        for i in range(count):
            ra = (start_ra + i * 45) % 360
            dec = -45 + i * 30
            targets.append(
                make_target(targetid=i + 1, ra=ra, dec=dec, merit=100 - i, ss_min=300)
            )
        return targets

    return _make_many


@pytest.fixture
def queue_get_from_list():
    """
    Fixture returning a helper to set scheduler.queue.get to pop entries
    from a provided list. Returns the recorded positions list if tracking
    is enabled.
    """

    def _set_queue_get(scheduler, targets, track_positions=False):
        call_count = {"count": 0}
        positions = []

        def getter(ra, dec, utime):
            if track_positions:
                positions.append((ra, dec))
            if call_count["count"] < len(targets):
                res = targets[call_count["count"]]
                res.done = False
                call_count["count"] += 1
                return res
            return None

        scheduler.queue.get = getter
        return positions

    return _set_queue_get


@pytest.fixture
def scheduler_2022_100_len2(mock_queue):
    plan = Plan()
    begin = datetime(2022, 4, 10, tzinfo=timezone.utc)
    end = begin + timedelta(days=2)
    return DumbQueueScheduler(queue=mock_queue, plan=plan, begin=begin, end=end)


# TOO-related fixtures for reducing test code duplication


@pytest.fixture
def standard_too_params():
    """Standard TOO parameters used in many tests."""
    return {
        "obsid": 1000001,
        "ra": 180.0,
        "dec": 45.0,
        "merit": 10000.0,
        "exptime": 3600,
        "name": "Test TOO",
    }


@pytest.fixture
def submitted_too(queue_ditl, standard_too_params):
    """A QueueDITL instance with a standard TOO already submitted."""
    return queue_ditl.submit_too(**standard_too_params)


@pytest.fixture
def low_merit_current_ppt(queue_ditl):
    """Create a current pointing object with low merit for TOO interrupt tests."""
    from conops import Pointing

    ppt = Pointing(
        config=queue_ditl.config,
        ra=0.0,
        dec=0.0,
        obsid=1,
        name="Current obs",
        merit=100.0,  # Lower merit than typical TOO
        exptime=1800,
    )
    queue_ditl.ppt = ppt
    return ppt


@pytest.fixture
def mock_pointing_visibility():
    """Mock patch for Pointing.visibility method."""
    with patch("conops.targets.pointing.Pointing.visibility") as mock_visibility:
        mock_visibility.return_value = None  # visibility() just populates windows
        yield mock_visibility


@pytest.fixture
def mock_pointing_visible():
    """Mock patch for Pointing.visible method."""
    with patch("conops.targets.pointing.Pointing.visible") as mock_visible:
        mock_visible.return_value = True  # Default to visible
        yield mock_visible


@pytest.fixture
def mock_terminate_ppt():
    """Mock patch for QueueDITL._terminate_ppt method."""
    with patch("conops.QueueDITL._terminate_ppt") as mock_terminate:
        yield mock_terminate


@pytest.fixture
def mock_fetch_new_ppt():
    """Mock patch for QueueDITL._fetch_new_ppt method."""
    with patch("conops.QueueDITL._fetch_new_ppt") as mock_fetch:
        yield mock_fetch


@pytest.fixture
def basic_too_request():
    """Create a basic TOORequest object with standard parameters."""
    from conops.ditl import TOORequest

    return TOORequest(
        obsid=1000001,
        ra=180.0,
        dec=45.0,
        merit=10000.0,
        exptime=3600,
        name="GRB 250101A",
    )


@pytest.fixture
def custom_too_request():
    """Create a TOORequest object with custom submit_time and executed=True."""
    from conops.ditl import TOORequest

    return TOORequest(
        obsid=1000002,
        ra=90.0,
        dec=-30.0,
        merit=5000.0,
        exptime=1800,
        name="Test TOO",
        submit_time=1234567890.0,
        executed=True,
    )


@pytest.fixture
def unix_timestamp_too(queue_ditl):
    """A QueueDITL instance with a TOO submitted using Unix timestamp."""
    submit_time = 1640995200.0  # 2022-01-01 00:00:00 UTC
    return queue_ditl.submit_too(
        obsid=1000002,
        ra=90.0,
        dec=-30.0,
        merit=5000.0,
        exptime=1800,
        name="Scheduled TOO",
        submit_time=submit_time,
    )


@pytest.fixture
def mock_too_interrupt_success(
    mock_pointing_visibility,
    mock_pointing_visible,
    mock_terminate_ppt,
    mock_fetch_new_ppt,
):
    """Combined mock setup for successful TOO interrupt tests."""
    mock_pointing_visibility.return_value = None  # visibility() just populates windows
    mock_pointing_visible.return_value = True  # Target is visible
    return {
        "visibility": mock_pointing_visibility,
        "visible": mock_pointing_visible,
        "terminate": mock_terminate_ppt,
        "fetch": mock_fetch_new_ppt,
    }


@pytest.fixture
def mock_too_interrupt_no_current_obs(
    mock_pointing_visibility, mock_pointing_visible, mock_fetch_new_ppt
):
    """Combined mock setup for TOO interrupt tests with no current observation."""
    mock_pointing_visibility.return_value = None  # visibility() just populates windows
    mock_pointing_visible.return_value = True  # Target is visible
    return {
        "visibility": mock_pointing_visibility,
        "visible": mock_pointing_visible,
        "fetch": mock_fetch_new_ppt,
    }


@pytest.fixture
def queue_ditl_no_queue_log(mock_config, mock_ephem):
    """Create a QueueDITL instance with a queue that has no log initially."""
    with (
        patch("conops.Queue") as mock_queue_class,
        patch("conops.PassTimes") as mock_passtimes,
        patch("conops.ACS") as mock_acs_class,
    ):
        # Mock PassTimes
        mock_pt = Mock()
        mock_pt.passes = []
        mock_pt.get = Mock()
        mock_pt.check_pass_timing = Mock(
            return_value={"start_pass": None, "end_pass": False, "updated_pass": None}
        )
        mock_passtimes.return_value = mock_pt

        # Mock ACS
        mock_acs = Mock()
        mock_acs.ephem = mock_ephem
        mock_acs.slewing = False
        mock_acs.inpass = False
        mock_acs.saa = None
        mock_acs.pointing = Mock(return_value=(0.0, 0.0, 0.0, 0))
        mock_acs.enqueue_command = Mock()
        mock_acs.passrequests = mock_pt
        mock_acs.slew_dists = []
        mock_acs.last_slew = None
        # Set acsmode to a real ACSMode enum value for logging
        from conops import ACSMode

        mock_acs.acsmode = ACSMode.SCIENCE
        # Mock the helper methods used in _fetch_new_ppt
        mock_target_request = Mock()
        mock_target_request.next_vis = Mock(return_value=1000.0)
        mock_acs._create_target_request = Mock(return_value=mock_target_request)
        mock_acs._initialize_slew_positions = Mock(return_value=True)
        mock_acs._is_slew_valid = Mock(return_value=True)
        mock_acs._calculate_slew_timing = Mock(return_value=1000.0)
        mock_acs_class.return_value = mock_acs

        # Mock solar panel in config (not ACS)
        mock_config.solar_panel.illumination_and_power = Mock(return_value=(0.5, 100.0))

        # Mock Queue with no log initially
        mock_queue = Mock()
        mock_queue.get = Mock(return_value=None)
        mock_queue.log = None  # Explicitly set log to None
        mock_queue_class.return_value = mock_queue

        ditl = QueueDITL(config=mock_config, ephem=mock_ephem, queue=mock_queue)
        ditl.acs = mock_acs

        return ditl


@pytest.fixture
def queue_ditl_acs_no_ephem(mock_config, mock_ephem):
    """Create a QueueDITL instance where ACS initially has no ephem."""
    with (
        patch("conops.Queue") as mock_queue_class,
        patch("conops.PassTimes") as mock_passtimes,
        patch("conops.ACS") as mock_acs_class,
    ):
        # Mock PassTimes
        mock_pt = Mock()
        mock_pt.passes = []
        mock_pt.get = Mock()
        mock_pt.check_pass_timing = Mock(
            return_value={"start_pass": None, "end_pass": False, "updated_pass": None}
        )
        mock_passtimes.return_value = mock_pt

        # Mock ACS with no ephem initially
        mock_acs = Mock()
        mock_acs.ephem = None  # ACS has no ephem initially
        mock_acs.slewing = False
        mock_acs.inpass = False
        mock_acs.saa = None
        mock_acs.pointing = Mock(return_value=(0.0, 0.0, 0.0, 0))
        mock_acs.enqueue_command = Mock()
        mock_acs.passrequests = mock_pt
        mock_acs.slew_dists = []
        mock_acs.last_slew = None
        # Set acsmode to a real ACSMode enum value for logging
        from conops import ACSMode

        mock_acs.acsmode = ACSMode.SCIENCE
        # Mock the helper methods used in _fetch_new_ppt
        mock_target_request = Mock()
        mock_target_request.next_vis = Mock(return_value=1000.0)
        mock_acs._create_target_request = Mock(return_value=mock_target_request)
        mock_acs._initialize_slew_positions = Mock(return_value=True)
        mock_acs._is_slew_valid = Mock(return_value=True)
        mock_acs._calculate_slew_timing = Mock(return_value=1000.0)
        mock_acs_class.return_value = mock_acs

        # Mock solar panel in config (not ACS)
        mock_config.solar_panel.illumination_and_power = Mock(return_value=(0.5, 100.0))

        # Mock Queue
        mock_queue = Mock()
        mock_queue.get = Mock(return_value=None)
        mock_queue_class.return_value = mock_queue

        ditl = QueueDITL(config=mock_config, ephem=mock_ephem, queue=mock_queue)
        ditl.acs = mock_acs

        return ditl



import time

import numpy as np
import rust_ephem
from pydantic import BaseModel, Field

from ..common import ics_date_conv, unixtime2date
from ..common.enums import AntennaType
from ..common.vector import radec2vec, rotvec, separation, vec2radec
from ..config import Constraint, GroundStationRegistry, MissionConfig
from ..config.constants import DTOR


class Pass(BaseModel):
    """A groundstation pass consisting of the dwell phase.

    The pass contains pointing profile information for ground station contacts.
    """

    model_config = {"arbitrary_types_allowed": True}

    # Core dependencies
    ephem: rust_ephem.Ephemeris | None = None
    config: MissionConfig | None = None

    # Pass metadata
    station: str
    begin: float
    length: float

    # What type of observation is this, a Ground Station Pass (GSP)
    obstype: str = "GSP"

    # Ground station pointing vectors (start/end of contact)
    gsstartra: float = 0.0
    gsstartdec: float = 0.0
    gsendra: float = 0.0
    gsenddec: float = 0.0

    # Recorded pointing profile during the pass dwell
    utime: list[float] = Field(default_factory=list)
    ra: list[float] = Field(default_factory=list)
    dec: list[float] = Field(default_factory=list)

    # Scheduling / status
    slewrequired: float = 0.0
    slewlate: float = 0.0
    obsid: int = 0xFFFF

    @property
    def end(self) -> float:
        assert self.length is not None, "Pass length must be set"
        return self.begin + self.length

    def __str__(self) -> str:
        """Return string of details on the pass"""
        return f"{unixtime2date(self.begin):18s}  {self.station:3s}  {self.length / 60.0:4.1f} mins"  #  {self.time_to_pass():12s}"

    def in_pass(self, utime: float) -> bool:
        if utime >= self.begin and utime <= self.end:
            return True
        else:
            return False

    def time_to_pass(self) -> str:
        """Return a string for how long it is until the next pass"""
        now = time.time()
        timetopass = (self.begin - now) / 60.0

        if timetopass < 60.0:
            timetopassstring = "%.0f mins " % timetopass
        else:
            timetopassstring = "%.1f hours" % (timetopass / 60.0)

        return timetopassstring

    def ra_dec(self, utime: float) -> tuple[float | None, float | None]:
        """Return the RA/Dec of the Spacecraft during a groundstation pass.
        Note: If utime is outside the pass, returns the earliest or latest
        RA/Dec in the pass."""

        # Find the closest time index
        idx = np.searchsorted(self.utime, utime)
        # searchsorted will return length if utime is beyond the end
        if idx >= len(self.utime):
            return self.ra[-1], self.dec[-1]
        return self.ra[idx], self.dec[idx]

    @staticmethod
    def apply_antenna_offset(
        ra: float, dec: float, azimuth_deg: float, elevation_deg: float
    ) -> tuple[float, float]:
        """Apply antenna pointing offset to RA/Dec.

        For a fixed antenna pointing at (azimuth, elevation) in the spacecraft body frame,
        calculate the adjusted RA/Dec that the antenna actually points toward.

        Args:
            ra: Original right ascension (degrees)
            dec: Original declination (degrees)
            azimuth_deg: Antenna azimuth offset (degrees, 0=forward, 90=right)
            elevation_deg: Antenna elevation offset (degrees, 0=horizon, 90=zenith, -90=nadir)

        Returns:
            Tuple of (adjusted_ra, adjusted_dec) in degrees
        """

        # Convert RA/Dec to unit vector
        sat_vec = radec2vec(ra * DTOR, dec * DTOR)

        # Apply rotation for azimuth (rotation around Z-axis/north pole)
        # Positive azimuth rotates right (east)
        # rotvec parameters: (axis_number, angle, vector) where axis: 1=x, 2=y, 3=z
        rotated = rotvec(3, azimuth_deg * DTOR, sat_vec)

        # Apply rotation for elevation (rotation around east-west axis = Y)
        # Positive elevation points up from horizon toward zenith
        # Negative elevation points down toward nadir
        rotated = rotvec(2, -elevation_deg * DTOR, rotated)

        # Convert back to RA/Dec
        adjusted_ra, adjusted_dec = np.degrees(vec2radec(rotated))

        return adjusted_ra, adjusted_dec

    def pointing_error(
        self,
        spacecraft_ra: float,
        spacecraft_dec: float,
        target_ra: float,
        target_dec: float,
    ) -> float:
        """Calculate pointing error between spacecraft pointing and target.

        Args:
            spacecraft_ra: Current spacecraft RA (degrees)
            spacecraft_dec: Current spacecraft Dec (degrees)
            target_ra: Target RA (degrees)
            target_dec: Target Dec (degrees)

        Returns:
            Angular separation in degrees
        """
        # Convert to radians

        return (
            separation(
                [spacecraft_ra * DTOR, spacecraft_dec * DTOR],
                [target_ra * DTOR, target_dec * DTOR],
            )
            / DTOR
        )

    def can_communicate(
        self, spacecraft_ra: float, spacecraft_dec: float, utime: float | None = None
    ) -> bool:
        """Check if communication is possible given spacecraft pointing.

        Args:
            spacecraft_ra: Current spacecraft RA (degrees)
            spacecraft_dec: Current spacecraft Dec (degrees)
            utime: Time during pass (optional, uses begin if None)

        Returns:
            True if communication is possible, False otherwise
        """
        if self.config is None or self.config.spacecraft_bus.communications is None:
            # No comms config, assume communication is always possible
            return True

        # Determine target pointing for this time
        if utime is None:
            utime = self.begin

        target_ra, target_dec = self.ra_dec(utime)
        if target_ra is None or target_dec is None:
            return False

        # Calculate pointing error
        error = self.pointing_error(
            spacecraft_ra, spacecraft_dec, target_ra, target_dec
        )

        # Check if within acceptable range
        return self.config.spacecraft_bus.communications.can_communicate(error)

    def get_data_rate(self, band: str, direction: str = "downlink") -> float:
        """Get data rate for this pass in the specified band.

        Args:
            band: Frequency band (e.g., "S", "X", "Ka")
            direction: "downlink" or "uplink"

        Returns:
            Data rate in Mbps, or 0.0 if band not supported
        """
        assert self.config is not None, "Config must be set for Pass class"
        if self.config.spacecraft_bus.communications is None:
            return 0.0

        if direction.lower() == "downlink":
            return self.config.spacecraft_bus.communications.get_downlink_rate(band)
        elif direction.lower() == "uplink":
            return self.config.spacecraft_bus.communications.get_uplink_rate(band)
        else:
            return 0.0

    def calculate_data_volume(self, band: str, direction: str = "downlink") -> float:
        """Calculate total data volume for this pass.

        Args:
            band: Frequency band (e.g., "S", "X", "Ka")
            direction: "downlink" or "uplink"

        Returns:
            Data volume in Megabits
        """
        assert self.config is not None, "Config must be set for Pass class"
        if self.length is None or self.config.spacecraft_bus.communications is None:
            return 0.0

        rate_mbps = self.get_data_rate(band, direction)
        return rate_mbps * self.length  # Mbps * seconds = Megabits

    def time_to_slew(self, utime: float, ra: float, dec: float) -> bool:
        """Determine whether to begin slewing for this pass.

        Calculates the slew time between current RA/Dec and the appropriate pointing of the pass.
        If on time, slews to pass start. If late, slews to where the pass currently is.
        Returns True when the pass time minus slew time is less than 60 seconds away.
        """
        assert self.ephem is not None, "Ephemeris must be set for Pass class"
        assert self.config is not None, "Config must be set for Pass class"

        # Determine target pointing: if we're late, target where pass currently is
        if utime >= self.begin:
            # We're late - target current pass position
            target_ra, target_dec = self.ra_dec(utime)
            if target_ra is None or target_dec is None:
                return False
        else:
            # On time - target pass start
            if not self.ra or not self.dec:
                return False
            target_ra, target_dec = self.ra[0], self.dec[0]

        # Using ACS config, calculate slew time
        if self.config.spacecraft_bus.attitude_control is not None:
            slewdist, _ = self.config.spacecraft_bus.attitude_control.predict_slew(
                startra=ra,
                startdec=dec,
                endra=target_ra,
                enddec=target_dec,
            )
            slewtime = self.config.spacecraft_bus.attitude_control.slew_time(slewdist)
        else:
            raise ValueError("ACS config must be set to calculate slew time")
        # Determine if we need to start slewing now
        time_until_slew = (self.begin - slewtime) - utime

        # If we are within 2 ephem steps of needing to slew, return True
        # FIXME: Is this buffer necessary?
        if time_until_slew <= self.ephem.step_size * 2:
            return True
        else:
            return False


class PassTimes:
    """Like the swift PassTimes class, except we calculate our passes based on epheris and groundstation location"""

    passes: list[Pass]
    constraint: Constraint
    ephem: rust_ephem.Ephemeris
    ground_stations: GroundStationRegistry
    config: MissionConfig

    def __init__(
        self,
        config: MissionConfig,
    ):
        self.constraint = config.constraint
        assert self.constraint is not None, "Constraint must be set for PassTimes class"
        assert self.constraint.ephem is not None, (
            "Ephemeris must be set for PassTimes class"
        )
        self.ephem = self.constraint.ephem

        self.config = config
        self.passes = []
        self.length = 1

        # Ground stations registry from config
        if config.ground_stations is None:
            self.ground_stations = GroundStationRegistry.default()
        else:
            self.ground_stations = config.ground_stations

        # What makes a good pass
        self.minelev = 10.0
        self.minlen = 8 * 60  # 10 mins
        self.schedule_chance = 1.0  # base chance of getting a pass

    def __getitem__(self, number: int) -> Pass:
        return self.passes[number]

    def __len__(self) -> int:
        return len(self.passes)

    def next_pass(self, utime: float) -> Pass | None:
        for gspass in self.passes:
            if utime < gspass.begin:
                return gspass
        return None

    def current_pass(self, utime: float) -> Pass | None:
        """Get the current active pass being tracked."""
        for gspass in self.passes:
            if gspass.in_pass(utime):
                return gspass
        return None

    def request_passes(self, req_gsnum: int, gsprob: float = 0.9) -> list[Pass]:
        """Request passes at a particular rate, including random probability of scheduling"""
        mean_between = 86400 / req_gsnum
        sched = list()
        last = 0.0
        for gspass in self.passes:
            if gspass.begin - last > mean_between and np.random.random() < gsprob:
                sched.extend([gspass])
                last = sched[-1].begin
        return sched

    def get(self, year: int, day: int, length: int = 1) -> None:
        """Calculate the passes using rust_ephem GroundEphemeris for vectorized operations."""
        ustart = ics_date_conv(f"{year}-{day:03d}-00:00:00")
        assert self.ephem is not None, "Ephemeris is not set"

        # Use binary search instead of np.where for finding start index
        # Prefer adapter datetimes if available, otherwise use Time.unix
        timestamp_unix = np.array([dt.timestamp() for dt in self.ephem.timestamp])
        startindex = int(np.searchsorted(timestamp_unix, ustart))

        # Calculate end index
        num_steps = int(86400 * length / self.ephem.step_size)
        endindex = min(startindex + num_steps, len(timestamp_unix))

        # Get timestamps as datetime objects for GroundEphemeris
        timestamps = self.ephem.timestamp[startindex:endindex]
        begin_time = timestamps[0]
        end_time = timestamps[-1]

        # Process each ground station
        for station in self.ground_stations.stations:
            # Create GroundEphemeris for this station (vectorized ground station ephemeris)

            gs_ephem = rust_ephem.GroundEphemeris(
                latitude=station.latitude_deg,
                longitude=station.longitude_deg,
                height=station.elevation_m,
                begin=begin_time,
                end=end_time,
                step_size=self.ephem.step_size,
            )

            # Calculate satellite RA/Dec as seen from ground station
            sat_ra, sat_dec = np.degrees(
                vec2radec(
                    (
                        self.ephem.gcrs_pv.position[startindex:endindex]
                        - gs_ephem.gcrs_pv.position
                    ).T
                )
            )

            # Fast vectorized approach: compute Earth limb constraint directly
            # The Earth limb constraint checks if the angle from the observer to the target
            # passes through Earth (i.e., target is below the horizon + min_angle)

            # Get ground station position in GCRS
            gs_pos = gs_ephem.gcrs_pv.position  # Shape: (N, 3)

            # Vector from ground station to satellite (target direction)
            gs_to_sat = (
                self.ephem.gcrs_pv.position[startindex:endindex] - gs_pos
            )  # Shape: (N, 3)

            # Normalize to get unit vector toward satellite
            gs_to_sat_dist = np.linalg.norm(gs_to_sat, axis=1, keepdims=True)
            gs_to_sat_unit = gs_to_sat / gs_to_sat_dist

            # Vector from Earth center to ground station
            earth_to_gs = gs_pos  # GCRS origin is Earth center
            earth_to_gs_dist = np.linalg.norm(earth_to_gs, axis=1, keepdims=True)
            earth_to_gs_unit = earth_to_gs / earth_to_gs_dist

            # Angle between "up" (away from Earth center) and target direction
            # cos(angle) = dot(earth_to_gs_unit, gs_to_sat_unit)
            cos_angle = np.sum(earth_to_gs_unit * gs_to_sat_unit, axis=1)

            # Calculate elevation above local horizon
            elevation_angle = np.degrees(np.arcsin(cos_angle))

            # Find passes using elevation threshold
            min_elev = (
                station.min_elevation_deg
                if hasattr(station, "min_elevation_deg")
                else self.minelev
            )

            # Target is visible if elevation > min_elev
            is_visible = elevation_angle > min_elev

            # Find pass boundaries by detecting transitions
            transitions = np.diff(is_visible.astype(int))
            pass_starts = np.where(transitions == 1)[0] + 1
            pass_ends = np.where(transitions == -1)[0] + 1

            # Handle edge cases
            if is_visible[0]:
                pass_starts = np.concatenate([[0], pass_starts])
            if is_visible[-1]:
                pass_ends = np.concatenate([pass_ends, [len(is_visible)]])

            # Create Pass objects
            for start_idx, end_idx in zip(pass_starts, pass_ends):
                global_start_idx = startindex + start_idx
                global_end_idx = (
                    startindex + end_idx
                )  # end_idx is already the first point below threshold

                passstart = timestamp_unix[global_start_idx]
                passend = timestamp_unix[global_end_idx]
                passlen = passend - passstart

                # Only consider passes that meet minimum length
                if passlen >= self.minlen:
                    # Combine global and station-specific schedule probabilities
                    combined_prob = self.schedule_chance * getattr(
                        station, "schedule_probability", 1.0
                    )
                    rand_val = np.random.random()
                    if rand_val <= combined_prob:
                        # Schedule this pass if the dice roll allows it
                        gspass = Pass(
                            config=self.config,
                            ephem=self.ephem,
                            station=station.code,
                            begin=passstart,
                            gsstartra=sat_ra[start_idx],
                            gsstartdec=sat_dec[start_idx],
                            gsendra=sat_ra[end_idx - 1],
                            gsenddec=sat_dec[end_idx - 1],
                            length=passlen,
                        )

                        # Record the path during the pass
                        gspass.utime = timestamp_unix[
                            startindex + start_idx : startindex + end_idx
                        ].tolist()
                        gspass.ra = sat_ra[start_idx:end_idx].tolist()
                        gspass.dec = sat_dec[start_idx:end_idx].tolist()

                        # Apply antenna pointing offset for fixed antennas
                        if (
                            gspass.config is not None
                            and gspass.config.spacecraft_bus.communications is not None
                            and gspass.config.spacecraft_bus.communications.antenna_pointing.antenna_type
                            == AntennaType.FIXED
                        ):
                            antenna = gspass.config.spacecraft_bus.communications.antenna_pointing
                            # Adjust start/end pointing
                            gspass.gsstartra, gspass.gsstartdec = (
                                Pass.apply_antenna_offset(
                                    gspass.gsstartra,
                                    gspass.gsstartdec,
                                    antenna.fixed_azimuth_deg,
                                    antenna.fixed_elevation_deg,
                                )
                            )
                            gspass.gsendra, gspass.gsenddec = Pass.apply_antenna_offset(
                                gspass.gsendra,
                                gspass.gsenddec,
                                antenna.fixed_azimuth_deg,
                                antenna.fixed_elevation_deg,
                            )
                            # Adjust full pointing profile
                            adjusted_pointing = [
                                Pass.apply_antenna_offset(
                                    ra_val,
                                    dec_val,
                                    antenna.fixed_azimuth_deg,
                                    antenna.fixed_elevation_deg,
                                )
                                for ra_val, dec_val in zip(gspass.ra, gspass.dec)
                            ]
                            gspass.ra = [pt[0] for pt in adjusted_pointing]
                            gspass.dec = [pt[1] for pt in adjusted_pointing]

                        self.passes.append(gspass)

        # Order the passes by time
        self.passes.sort(key=lambda x: x.begin, reverse=False)



"""Unit tests for spacecraft_bus module."""

import numpy as np

from conops import AttitudeControlSystem, PowerDraw, SpacecraftBus


class TestPowerDraw:
    """Tests for PowerDraw class."""

    def test_initialization_defaults_nominal_power(self, default_power_draw):
        """Test PowerDraw initializes with default nominal_power."""
        assert default_power_draw.nominal_power == 50

    def test_initialization_defaults_peak_power(self, default_power_draw):
        """Test PowerDraw initializes with default peak_power."""
        assert default_power_draw.peak_power == 300

    def test_initialization_defaults_power_mode(self, default_power_draw):
        """Test PowerDraw initializes with default power_mode."""
        assert default_power_draw.power_mode == {}

    def test_initialization_custom_nominal_power(self, custom_power_draw):
        """Test PowerDraw initializes with custom nominal_power."""
        assert custom_power_draw.nominal_power == 150

    def test_initialization_custom_peak_power(self, custom_power_draw):
        """Test PowerDraw initializes with custom peak_power."""
        assert custom_power_draw.peak_power == 250

    def test_initialization_custom_power_mode(self, custom_power_draw):
        """Test PowerDraw initializes with custom power_mode."""
        assert custom_power_draw.power_mode == {1: 175, 2: 225}

    def test_power_no_mode(self):
        """Test power() returns nominal_power when no mode specified."""
        pd = PowerDraw(nominal_power=180)
        assert pd.power() == 180

    def test_power_none_mode(self):
        """Test power() returns nominal_power when mode is None."""
        pd = PowerDraw(nominal_power=180)
        assert pd.power(None) == 180

    def test_power_with_mode_1(self):
        """Test power() returns mode-specific power for mode 1."""
        pd = PowerDraw(nominal_power=100, power_mode={1: 120, 2: 140})
        assert pd.power(1) == 120

    def test_power_with_mode_2(self):
        """Test power() returns mode-specific power for mode 2."""
        pd = PowerDraw(nominal_power=100, power_mode={1: 120, 2: 140})
        assert pd.power(2) == 140

    def test_power_undefined_mode(self):
        """Test power() returns nominal_power for undefined modes."""
        pd = PowerDraw(nominal_power=100, power_mode={1: 120})
        assert pd.power(99) == 100


class TestAttitudeControlSystem:
    """Tests for AttitudeControlSystem class."""

    def test_initialization_defaults_slew_acceleration(self, default_acs):
        """Test ACS initializes with default slew_acceleration."""
        assert default_acs.slew_acceleration == 0.5

    def test_initialization_defaults_max_slew_rate(self, default_acs):
        """Test ACS initializes with default max_slew_rate."""
        assert default_acs.max_slew_rate == 0.25

    def test_initialization_defaults_slew_accuracy(self, default_acs):
        """Test ACS initializes with default slew_accuracy."""
        assert default_acs.slew_accuracy == 0.01

    def test_initialization_defaults_settle_time(self, default_acs):
        """Test ACS initializes with default settle_time."""
        assert default_acs.settle_time == 120.0

    def test_initialization_custom_slew_acceleration(self, custom_acs):
        """Test ACS initializes with custom slew_acceleration."""
        assert custom_acs.slew_acceleration == 1.0

    def test_initialization_custom_max_slew_rate(self, custom_acs):
        """Test ACS initializes with custom max_slew_rate."""
        assert custom_acs.max_slew_rate == 0.5

    def test_initialization_custom_slew_accuracy(self, custom_acs):
        """Test ACS initializes with custom slew_accuracy."""
        assert custom_acs.slew_accuracy == 0.05

    def test_initialization_custom_settle_time(self, custom_acs):
        """Test ACS initializes with custom settle_time."""
        assert custom_acs.settle_time == 60.0

    def test_motion_time_zero_angle(self, default_acs):
        """Test motion_time returns 0 for zero angle."""
        assert default_acs.motion_time(0) == 0.0

    def test_motion_time_negative_angle(self, default_acs):
        """Test motion_time returns 0 for negative angles."""
        assert default_acs.motion_time(-10) == 0.0

    def test_motion_time_invalid_params_zero_acceleration(self):
        """Test motion_time returns 0 for zero acceleration."""
        acs = AttitudeControlSystem(slew_acceleration=0)
        assert acs.motion_time(10) == 0.0

    def test_motion_time_invalid_params_zero_rate(self):
        """Test motion_time returns 0 for zero max_slew_rate."""
        acs = AttitudeControlSystem(max_slew_rate=0)
        assert acs.motion_time(10) == 0.0

    def test_motion_time_triangular_profile(self):
        """Test motion_time for small angle (triangular velocity profile)."""
        # With a=0.5, vmax=0.25: t_accel=0.5, d_accel=0.0625
        # 2*d_accel = 0.125, so angles < 0.125 use triangular profile
        acs = AttitudeControlSystem(slew_acceleration=0.5, max_slew_rate=0.25)
        angle = 0.1  # Less than 2*d_accel
        motion_time = acs.motion_time(angle)
        # For triangular: t_peak = sqrt(angle/a), total = 2*t_peak
        expected = 2 * np.sqrt(angle / 0.5)
        assert abs(motion_time - expected) < 1e-6

    def test_motion_time_trapezoidal_profile(self):
        """Test motion_time for large angle (trapezoidal velocity profile)."""
        acs = AttitudeControlSystem(slew_acceleration=0.5, max_slew_rate=0.25)
        angle = 10.0  # Much larger than 2*d_accel
        motion_time = acs.motion_time(angle)
        # t_accel = 0.5, d_accel = 0.0625
        # d_cruise = 10 - 2*0.0625 = 9.875
        # t_cruise = 9.875 / 0.25 = 39.5
        # total = 2*0.5 + 39.5 = 40.5
        expected = 2 * 0.5 + 9.875 / 0.25
        assert abs(motion_time - expected) < 1e-6

    def test_s_of_t_zero_angle(self, default_acs):
        """Test s_of_t returns 0 for zero angle."""
        assert default_acs.s_of_t(0, 10) == 0.0

    def test_s_of_t_zero_time(self, default_acs):
        """Test s_of_t returns 0 for zero time."""
        assert default_acs.s_of_t(10, 0) == 0.0

    def test_s_of_t_negative_angle(self, default_acs):
        """Test s_of_t returns 0 for negative angle."""
        assert default_acs.s_of_t(-5, 10) == 0.0

    def test_s_of_t_invalid_params(self):
        """Test s_of_t fallback for invalid parameters."""
        acs = AttitudeControlSystem(slew_acceleration=0, max_slew_rate=0.25)
        # Should use fallback: min(max(0, t*vmax), angle)
        result = acs.s_of_t(10, 5)
        expected = min(5 * 0.25, 10)
        assert abs(result - expected) < 1e-6

    def test_s_of_t_triangular_acceleration(self):
        """Test s_of_t during acceleration phase of triangular profile."""
        acs = AttitudeControlSystem(slew_acceleration=0.5, max_slew_rate=0.25)
        angle = 0.1
        t_peak = np.sqrt(angle / 0.5)
        t = t_peak * 0.5  # Halfway through acceleration
        s = acs.s_of_t(angle, t)
        expected = 0.5 * 0.5 * t**2
        assert abs(s - expected) < 1e-6

    def test_s_of_t_triangular_deceleration(self):
        """Test s_of_t during deceleration phase of triangular profile."""
        acs = AttitudeControlSystem(slew_acceleration=0.5, max_slew_rate=0.25)
        angle = 0.1
        t_peak = np.sqrt(angle / 0.5)
        motion_time = 2 * t_peak
        t = motion_time * 0.75  # During deceleration
        s = acs.s_of_t(angle, t)
        expected = angle - 0.5 * 0.5 * (motion_time - t) ** 2
        assert abs(s - expected) < 1e-6

    def test_s_of_t_trapezoidal_acceleration(self):
        """Test s_of_t during acceleration phase of trapezoidal profile."""
        acs = AttitudeControlSystem(slew_acceleration=0.5, max_slew_rate=0.25)
        angle = 10.0
        t = 0.25  # During acceleration
        s = acs.s_of_t(angle, t)
        expected = 0.5 * 0.5 * t**2
        assert abs(s - expected) < 1e-6

    def test_s_of_t_trapezoidal_cruise(self):
        """Test s_of_t during cruise phase of trapezoidal profile."""
        acs = AttitudeControlSystem(slew_acceleration=0.5, max_slew_rate=0.25)
        angle = 10.0
        t_accel = 0.5
        d_accel = 0.0625
        t = 5.0  # During cruise
        s = acs.s_of_t(angle, t)
        expected = d_accel + 0.25 * (t - t_accel)
        assert abs(s - expected) < 1e-6

    def test_s_of_t_trapezoidal_deceleration(self):
        """Test s_of_t during deceleration phase of trapezoidal profile."""
        acs = AttitudeControlSystem(slew_acceleration=0.5, max_slew_rate=0.25)
        angle = 10.0
        t_accel = 0.5
        d_accel = 0.0625
        d_cruise = angle - 2 * d_accel
        t_cruise = d_cruise / 0.25
        t = t_accel + t_cruise + 0.1  # During deceleration
        s = acs.s_of_t(angle, t)
        t_dec = t - (t_accel + t_cruise)
        expected = d_accel + d_cruise + 0.25 * t_dec - 0.5 * 0.5 * t_dec**2
        assert abs(s - expected) < 1e-6

    def test_s_of_t_after_motion_complete(self):
        """Test s_of_t returns full angle after motion is complete."""
        acs = AttitudeControlSystem(slew_acceleration=0.5, max_slew_rate=0.25)
        angle = 10.0
        motion_time = acs.motion_time(angle)
        s = acs.s_of_t(angle, motion_time + 100)
        assert abs(s - angle) < 1e-6

    def test_slew_time(self):
        """Test slew_time includes motion time and settle time."""
        acs = AttitudeControlSystem(slew_acceleration=0.5, settle_time=60.0)
        angle = 10.0
        motion_time = acs.motion_time(angle)
        total_time = acs.slew_time(angle)
        assert abs(total_time - (motion_time + 60.0)) < 1e-6

    def test_slew_time_zero_angle(self, default_acs):
        """Test slew_time returns 0 for zero angle."""
        assert default_acs.slew_time(0) == 0.0

    def test_slew_time_negative_angle(self, default_acs):
        """Test slew_time returns 0 for negative angles."""
        assert default_acs.slew_time(-5) == 0.0

    def test_predict_slew_same_position_distance(self, default_acs):
        """Test predict_slew distance with identical start and end positions."""
        slewdist, slewpath = default_acs.predict_slew(0, 0, 0, 0)
        assert abs(slewdist) < 1e-6

    def test_predict_slew_same_position_path_length_ra(self, default_acs):
        """Test predict_slew path length for ra with identical positions."""
        slewdist, slewpath = default_acs.predict_slew(0, 0, 0, 0)
        ra_path, dec_path = slewpath
        # great_circle returns steps+2 points (includes start and end)
        assert len(ra_path) == 22  # 20 steps + start + end

    def test_predict_slew_same_position_path_length_dec(self, default_acs):
        """Test predict_slew path length for dec with identical positions."""
        slewdist, slewpath = default_acs.predict_slew(0, 0, 0, 0)
        ra_path, dec_path = slewpath
        # great_circle returns steps+2 points (includes start and end)
        assert len(dec_path) == 22

    def test_predict_slew_different_positions_distance(self, default_acs):
        """Test predict_slew calculates distance correctly."""
        # Simple case: 90 degree separation along equator
        slewdist, slewpath = default_acs.predict_slew(0, 0, 90, 0)
        assert abs(slewdist - 90.0) < 0.1  # Should be ~90 degrees

    def test_predict_slew_different_positions_path_length_ra(self, default_acs):
        """Test predict_slew path length for ra."""
        # Simple case: 90 degree separation along equator
        slewdist, slewpath = default_acs.predict_slew(0, 0, 90, 0)
        ra_path, dec_path = slewpath
        # great_circle returns steps+2 points (includes start and end)
        assert len(ra_path) == 22  # 20 steps + start + end

    def test_predict_slew_different_positions_path_length_dec(self, default_acs):
        """Test predict_slew path length for dec."""
        # Simple case: 90 degree separation along equator
        slewdist, slewpath = default_acs.predict_slew(0, 0, 90, 0)
        ra_path, dec_path = slewpath
        # great_circle returns steps+2 points (includes start and end)
        assert len(dec_path) == 22

    def test_predict_slew_different_positions_start_ra(self, default_acs):
        """Test predict_slew path starts at correct ra."""
        # Simple case: 90 degree separation along equator
        slewdist, slewpath = default_acs.predict_slew(0, 0, 90, 0)
        ra_path, dec_path = slewpath
        # Check path starts and ends at correct positions
        assert abs(ra_path[0] - 0) < 1e-6

    def test_predict_slew_different_positions_start_dec(self, default_acs):
        """Test predict_slew path starts at correct dec."""
        # Simple case: 90 degree separation along equator
        slewdist, slewpath = default_acs.predict_slew(0, 0, 90, 0)
        ra_path, dec_path = slewpath
        # Check path starts and ends at correct positions
        assert abs(dec_path[0] - 0) < 1e-6

    def test_predict_slew_different_positions_end_ra(self, default_acs):
        """Test predict_slew path ends at correct ra."""
        # Simple case: 90 degree separation along equator
        slewdist, slewpath = default_acs.predict_slew(0, 0, 90, 0)
        ra_path, dec_path = slewpath
        # Check path starts and ends at correct positions
        assert abs(ra_path[-1] - 90) < 1e-6

    def test_predict_slew_different_positions_end_dec(self, default_acs):
        """Test predict_slew path ends at correct dec."""
        # Simple case: 90 degree separation along equator
        slewdist, slewpath = default_acs.predict_slew(0, 0, 90, 0)
        ra_path, dec_path = slewpath
        # Check path starts and ends at correct positions
        assert abs(dec_path[-1] - 0) < 1e-6

    def test_predict_slew_custom_steps_path_length_ra(self, default_acs):
        """Test predict_slew with custom steps path length for ra."""
        slewdist, slewpath = default_acs.predict_slew(0, 0, 45, 45, steps=10)
        ra_path, dec_path = slewpath
        # great_circle returns steps+2 points (includes start and end)
        assert len(ra_path) == 12  # 10 steps + start + end

    def test_predict_slew_custom_steps_path_length_dec(self, default_acs):
        """Test predict_slew with custom steps path length for dec."""
        slewdist, slewpath = default_acs.predict_slew(0, 0, 45, 45, steps=10)
        ra_path, dec_path = slewpath
        # great_circle returns steps+2 points (includes start and end)
        assert len(dec_path) == 12

    def test_predict_slew_across_meridian_distance_positive(self, default_acs):
        """Test predict_slew across meridian has positive distance."""
        slewdist, slewpath = default_acs.predict_slew(350, 0, 10, 0)
        assert slewdist > 0  # Should find the shorter path

    def test_predict_slew_across_meridian_distance_less_than_180(self, default_acs):
        """Test predict_slew across meridian distance less than 180."""
        slewdist, slewpath = default_acs.predict_slew(350, 0, 10, 0)
        assert slewdist < 180  # Should not go the long way


class TestSpacecraftBus:
    """Tests for SpacecraftBus class."""

    def test_initialization_defaults_name(self, default_bus):
        """Test SpacecraftBus initializes with default name."""
        assert default_bus.name == "Default Bus"

    def test_initialization_defaults_power_draw_type(self, default_bus):
        """Test SpacecraftBus initializes with PowerDraw instance."""
        assert isinstance(default_bus.power_draw, PowerDraw)

    def test_initialization_defaults_attitude_control_type(self, default_bus):
        """Test SpacecraftBus initializes with AttitudeControlSystem instance."""
        assert isinstance(default_bus.attitude_control, AttitudeControlSystem)

    def test_initialization_custom_name(self, custom_bus):
        """Test SpacecraftBus initializes with custom name."""
        assert custom_bus.name == "Custom Bus"

    def test_initialization_custom_power_draw_nominal_power(self, custom_bus):
        """Test SpacecraftBus initializes with custom power_draw nominal_power."""
        assert custom_bus.power_draw.nominal_power == 150

    def test_initialization_custom_attitude_control_slew_acceleration(self, custom_bus):
        """Test SpacecraftBus initializes with custom attitude_control slew_acceleration."""
        assert custom_bus.attitude_control.slew_acceleration == 1.0

    def test_power_delegates_to_power_draw_no_mode(self):
        """Test SpacecraftBus.power() delegates to PowerDraw for no mode."""
        pd = PowerDraw(nominal_power=100, power_mode={1: 120})
        bus = SpacecraftBus(power_draw=pd)
        assert bus.power() == 100

    def test_power_delegates_to_power_draw_mode_1(self):
        """Test SpacecraftBus.power() delegates to PowerDraw for mode 1."""
        pd = PowerDraw(nominal_power=100, power_mode={1: 120})
        bus = SpacecraftBus(power_draw=pd)
        assert bus.power(1) == 120

    def test_power_delegates_to_power_draw_undefined_mode(self):
        """Test SpacecraftBus.power() delegates to PowerDraw for undefined mode."""
        pd = PowerDraw(nominal_power=100, power_mode={1: 120})
        bus = SpacecraftBus(power_draw=pd)
        assert bus.power(99) == 100


class TestSpacecraftBusEclipse:
    """Test eclipse-aware power consumption for spacecraft bus."""

    def test_bus_with_heater_eclipse_sunlight(self):
        """Test bus power with heater in sunlight."""
        from conops import Heater

        bus = SpacecraftBus(
            name="Test Bus",
            power_draw=PowerDraw(nominal_power=150.0),
            heater=Heater(
                name="Bus Heater",
                power_draw=PowerDraw(nominal_power=15.0, eclipse_power=35.0),
            ),
        )

        # Sunlight: base + heater
        assert bus.power(in_eclipse=False) == 165.0

    def test_bus_with_heater_eclipse_eclipse(self):
        """Test bus power with heater in eclipse."""
        from conops import Heater

        bus = SpacecraftBus(
            name="Test Bus",
            power_draw=PowerDraw(nominal_power=150.0),
            heater=Heater(
                name="Bus Heater",
                power_draw=PowerDraw(nominal_power=15.0, eclipse_power=35.0),
            ),
        )

        # Eclipse: base + higher heater power
        assert bus.power(in_eclipse=True) == 185.0

    def test_bus_base_power_with_eclipse_sunlight(self):
        """Test bus base power draw in sunlight."""
        bus = SpacecraftBus(
            name="Detector Bus",
            power_draw=PowerDraw(
                nominal_power=200.0,
                eclipse_power=210.0,  # Slightly higher in eclipse
            ),
        )

        assert bus.power(in_eclipse=False) == 200.0

    def test_bus_base_power_with_eclipse_eclipse(self):
        """Test bus base power draw in eclipse."""
        bus = SpacecraftBus(
            name="Detector Bus",
            power_draw=PowerDraw(
                nominal_power=200.0,
                eclipse_power=210.0,  # Slightly higher in eclipse
            ),
        )

        assert bus.power(in_eclipse=True) == 210.0

    def test_bus_full_eclipse_configuration_nominal_sunlight(self):
        """Test bus with both base and heater eclipse power in nominal sunlight."""
        from conops import Heater

        bus = SpacecraftBus(
            name="Science Bus",
            power_draw=PowerDraw(
                nominal_power=180.0,
                eclipse_power=190.0,
                power_mode={2: 250.0},
                eclipse_power_mode={2: 270.0},
            ),
            heater=Heater(
                name="Bus Heater",
                power_draw=PowerDraw(
                    nominal_power=20.0,
                    eclipse_power=50.0,
                    power_mode={2: 25.0},
                    eclipse_power_mode={2: 55.0},
                ),
            ),
        )

        # Nominal mode
        assert bus.power(in_eclipse=False) == 200.0  # 180 + 20

    def test_bus_full_eclipse_configuration_nominal_eclipse(self):
        """Test bus with both base and heater eclipse power in nominal eclipse."""
        from conops import Heater

        bus = SpacecraftBus(
            name="Science Bus",
            power_draw=PowerDraw(
                nominal_power=180.0,
                eclipse_power=190.0,
                power_mode={2: 250.0},
                eclipse_power_mode={2: 270.0},
            ),
            heater=Heater(
                name="Bus Heater",
                power_draw=PowerDraw(
                    nominal_power=20.0,
                    eclipse_power=50.0,
                    power_mode={2: 25.0},
                    eclipse_power_mode={2: 55.0},
                ),
            ),
        )

        # Nominal mode
        assert bus.power(in_eclipse=True) == 240.0  # 190 + 50

    def test_bus_full_eclipse_configuration_mode_2_sunlight(self):
        """Test bus with both base and heater eclipse power in mode 2 sunlight."""
        from conops import Heater

        bus = SpacecraftBus(
            name="Science Bus",
            power_draw=PowerDraw(
                nominal_power=180.0,
                eclipse_power=190.0,
                power_mode={2: 250.0},
                eclipse_power_mode={2: 270.0},
            ),
            heater=Heater(
                name="Bus Heater",
                power_draw=PowerDraw(
                    nominal_power=20.0,
                    eclipse_power=50.0,
                    power_mode={2: 25.0},
                    eclipse_power_mode={2: 55.0},
                ),
            ),
        )

        # Mode 2
        assert bus.power(mode=2, in_eclipse=False) == 275.0  # 250 + 25

    def test_bus_full_eclipse_configuration_mode_2_eclipse(self):
        """Test bus with both base and heater eclipse power in mode 2 eclipse."""
        from conops import Heater

        bus = SpacecraftBus(
            name="Science Bus",
            power_draw=PowerDraw(
                nominal_power=180.0,
                eclipse_power=190.0,
                power_mode={2: 250.0},
                eclipse_power_mode={2: 270.0},
            ),
            heater=Heater(
                name="Bus Heater",
                power_draw=PowerDraw(
                    nominal_power=20.0,
                    eclipse_power=50.0,
                    power_mode={2: 25.0},
                    eclipse_power_mode={2: 55.0},
                ),
            ),
        )

        # Mode 2
        assert bus.power(mode=2, in_eclipse=True) == 325.0  # 270 + 55

    def test_bus_no_heater_eclipse_sunlight(self):
        """Test bus without heater in sunlight."""
        bus = SpacecraftBus(
            name="Simple Bus",
            power_draw=PowerDraw(nominal_power=175.0, eclipse_power=185.0),
        )

        assert bus.power(in_eclipse=False) == 175.0

    def test_bus_no_heater_eclipse_eclipse(self):
        """Test bus without heater in eclipse."""
        bus = SpacecraftBus(
            name="Simple Bus",
            power_draw=PowerDraw(nominal_power=175.0, eclipse_power=185.0),
        )

        assert bus.power(in_eclipse=True) == 185.0

    def test_bus_eclipse_backward_compatible(self):
        """Test that not passing in_eclipse works (defaults to False)."""
        from conops import Heater

        bus = SpacecraftBus(
            power_draw=PowerDraw(nominal_power=150.0, eclipse_power=200.0),
            heater=Heater(
                name="Heater",
                power_draw=PowerDraw(nominal_power=10.0, eclipse_power=30.0),
            ),
        )

        # Default should be sunlight (in_eclipse=False)
        assert bus.power() == 160.0  # 150 + 10 (not eclipse values)



"""Test fixtures for passes subsystem tests."""

from datetime import datetime, timedelta, timezone
from unittest.mock import Mock

import numpy as np
import pytest
from rust_ephem import TLEEphemeris

from conops import Constraint, MissionConfig, Pass
from conops.config import (
    AttitudeControlSystem,
    Battery,
    GroundStationRegistry,
    Payload,
    SolarPanelSet,
    SpacecraftBus,
)


class MockEphemeris:
    """Mock ephemeris for testing."""

    def __init__(self, step_size=60.0, num_points=1440):
        self.step_size = step_size
        self.num_points = num_points

        # Create timestamps for one day
        base_time = datetime(2018, 1, 1, tzinfo=timezone.utc)
        self.timestamp = [
            base_time + timedelta(seconds=i * step_size) for i in range(num_points)
        ]

        # Create mock gcrs_pv and itrs_pv for position data
        self.gcrs_pv = Mock()
        self.itrs_pv = Mock()

        # Create position arrays (simplified orbital positions)
        positions = []
        for i in range(num_points):
            angle = 2 * np.pi * i / 1440  # One orbit per day
            x = 6900 * np.cos(angle)  # Earth radius + 550 km
            y = 6900 * np.sin(angle)
            z = 0
            positions.append([x, y, z])

        self.gcrs_pv.position = np.array(positions)
        self.itrs_pv.position = np.array(positions)

    def index(self, time):
        """Mock index method."""
        return 0


@pytest.fixture
def mock_ephem(create_ephem):
    """Create a simple real TLE ephemeris for tests.

    Tests previously used a MockEphemeris that fails Pydantic instance checks; use a
    short TLEEphemeris instead so model validation passes and tests can rely on
    actual ephemeris attributes like .timestamp, .gcrs_pv, etc.
    """
    # Use a short period ephemeris (15 minutes) for speed
    begin = datetime(2025, 8, 15, 0, 0, 0, tzinfo=timezone.utc)
    end = datetime(2025, 8, 15, 0, 15, 0, tzinfo=timezone.utc)
    return create_ephem(begin, end, step_size=60)


@pytest.fixture
def mock_constraint(mock_ephem):
    """Create a mock constraint."""
    constraint = Mock(spec=Constraint)
    constraint.ephem = mock_ephem
    return constraint


@pytest.fixture
def mock_acs_config():
    """Create a mock ACS config."""
    config = Mock()
    config.predict_slew = Mock(return_value=(10.0, np.array([[0, 0, 1]])))
    config.slew_time = Mock(return_value=10.0)  # Add slew_time method
    return config


@pytest.fixture
def mock_config(mock_constraint, mock_acs_config):
    """Create a Config object for testing."""
    # Create minimal required components
    spacecraft_bus = SpacecraftBus(
        attitude_control=AttitudeControlSystem(),
        communications=None,  # Will be set in tests that need it
    )

    # Create minimal payload, battery, solar_panel
    payload = Payload(instruments=[])
    battery = Battery(capacity_wh=1000, max_depth_of_discharge=0.8)
    solar_panel = SolarPanelSet(panels=[])

    config = MissionConfig(
        spacecraft_bus=spacecraft_bus,
        solar_panel=solar_panel,
        payload=payload,
        battery=battery,
        constraint=mock_constraint,
        ground_stations=GroundStationRegistry.default(),
    )
    return config


@pytest.fixture
def basic_pass_mock(mock_config, create_ephem):
    """Create a basic Pass instance."""
    ephem = create_ephem(1514764800.0, 1514764800.0 + 480.0)
    p = Pass(
        config=mock_config,
        ephem=ephem,
        station="SGS",
        begin=1514764800.0,
        length=480.0,
        gsstartra=10.0,
        gsstartdec=20.0,
        gsendra=15.0,
        gsenddec=25.0,
    )
    return p


@pytest.fixture
def mock_ephemeris_100():
    return MockEphemeris(step_size=60.0, num_points=100)


@pytest.fixture
def base_begin():
    return 1514764800.0


@pytest.fixture
def default_length():
    return 480.0


@pytest.fixture
def two_step_utime():
    return [1514764800.0, 1514764900.0]


@pytest.fixture
def three_step_utime():
    return [1514764800.0, 1514764900.0, 1514765000.0]


@pytest.fixture
def small_ra():
    return [10.0, 12.0]


@pytest.fixture
def small_dec():
    return [20.0, 22.0]


@pytest.fixture
def start_ra():
    return [15.0, 16.0]


@pytest.fixture
def start_dec():
    return [25.0, 26.0]


@pytest.fixture
def tle_path():
    return "examples/example.tle"


@pytest.fixture
def create_ephem(tle_path):
    def _create(begin, end, step_size=60):
        # Convert unix timestamp floats to timezone-aware datetimes if needed
        if isinstance(begin, (int, float)):
            begin = datetime.fromtimestamp(begin, tz=timezone.utc)
        if isinstance(end, (int, float)):
            end = datetime.fromtimestamp(end, tz=timezone.utc)
        return TLEEphemeris(tle=tle_path, begin=begin, end=end, step_size=step_size)

    return _create


@pytest.fixture
def acs_mock():
    acs = Mock()
    acs.predict_slew.return_value = (0.1, None)
    acs.slew_time.return_value = 100.0
    return acs


@pytest.fixture
def basic_pass_factory(mock_config, create_ephem):
    def _create(
        station="SGS",
        begin=1514764800.0,
        length=480.0,
    ):
        ephem = create_ephem(begin, begin + length)
        return Pass(
            config=mock_config,
            ephem=ephem,
            station=station,
            begin=begin,
            length=length,
            gsstartra=10.0,
            gsstartdec=20.0,
            gsendra=15.0,
            gsenddec=25.0,
        )

    return _create


@pytest.fixture
def basic_pass(basic_pass_factory):
    return basic_pass_factory()



"""Tests for conops.slew module."""

from unittest.mock import Mock

import numpy as np
import pytest

from conops import Slew


class TestSlewInit:
    """Test Slew initialization."""

    @pytest.mark.parametrize(
        "attr,value",
        [
            ("constraint", "constraint"),
            ("ephem", "ephem"),
            ("acs_config", "acs_config"),
            ("slewtime", 0),
            ("slewdist", 0),
            ("slewstart", 0),
            ("slewend", 0),
            ("slewrequest", 0),
            ("startra", 0),
            ("startdec", 0),
            ("endra", 0),
            ("enddec", 0),
            ("obsid", 0),
            ("mode", 0),
            ("at", None),
            ("obstype", "PPT"),
        ],
    )
    def test_slew_init(self, slew, attr, value, constraint, ephem, acs_config):
        if attr in ["constraint", "ephem", "acs_config"]:
            # Assuming fixtures provide the expected values
            expected = locals()[attr]
        else:
            expected = value
        assert getattr(slew, attr) == expected

    def test_slew_init_missing_constraint(self, acs_config):
        mock_config = Mock()
        mock_config.constraint = None
        mock_config.spacecraft_bus = Mock()
        mock_config.spacecraft_bus.attitude_control = acs_config
        with pytest.raises(AssertionError, match="Constraint must be set"):
            Slew(config=mock_config)

    def test_slew_init_missing_ephemeris(self, constraint, acs_config):
        constraint.ephem = None
        mock_config = Mock()
        mock_config.constraint = constraint
        mock_config.spacecraft_bus = Mock()
        mock_config.spacecraft_bus.attitude_control = acs_config
        with pytest.raises(AssertionError, match="Ephemeris must be set"):
            Slew(config=mock_config)

    def test_slew_init_missing_acs_config(self, constraint):
        mock_config = Mock()
        mock_config.constraint = constraint
        mock_config.spacecraft_bus = Mock()
        mock_config.spacecraft_bus.attitude_control = None
        with pytest.raises(AssertionError, match="ACS config must be set"):
            Slew(config=mock_config)


class TestSlewStr:
    """Test Slew string representation."""

    def test_slew_str_contains_slew_from(self, slew_with_positions):
        str_repr = str(slew_with_positions)
        assert "Slew from" in str_repr

    def test_slew_str_contains_45_000(self, slew_with_positions):
        str_repr = str(slew_with_positions)
        assert "45.000" in str_repr

    def test_slew_str_contains_30_000(self, slew_with_positions):
        str_repr = str(slew_with_positions)
        assert "30.000" in str_repr

    def test_slew_str_contains_90_0(self, slew_with_positions):
        str_repr = str(slew_with_positions)
        assert "90.0" in str_repr


class TestIsSlewing:
    """Test is_slewing method."""

    def test_is_slewing_at_1700000050_true(self, slew_slewing):
        assert slew_slewing.is_slewing(1700000050.0) is True

    def test_is_slewing_at_1699999999_false(self, slew_slewing):
        assert slew_slewing.is_slewing(1699999999.0) is False

    def test_is_slewing_at_1700000101_false(self, slew_slewing):
        assert slew_slewing.is_slewing(1700000101.0) is False

    def test_is_slewing_at_1700000000_true(self, slew_slewing):
        assert slew_slewing.is_slewing(1700000000.0) is True

    def test_is_slewing_at_1700000100_false(self, slew_slewing):
        assert slew_slewing.is_slewing(1700000100.0) is False


class TestRaDec:
    """Test ra_dec method."""

    def test_ra_dec_returns_ra_float(self, slew_ra_dec):
        ra, dec = slew_ra_dec.ra_dec(1700000000.0)
        assert isinstance(ra, (float, np.floating))

    def test_ra_dec_returns_dec_float(self, slew_ra_dec):
        ra, dec = slew_ra_dec.ra_dec(1700000000.0)
        assert isinstance(dec, (float, np.floating))


class TestSlewRaDec:
    """Test slew_ra_dec method."""

    def test_slew_ra_dec_returns_ra_float(self, slew_setup):
        ra, dec = slew_setup.slew_ra_dec(1700000000.0)
        assert isinstance(ra, (float, np.floating))

    def test_slew_ra_dec_returns_dec_float(self, slew_setup):
        ra, dec = slew_setup.slew_ra_dec(1700000000.0)
        assert isinstance(dec, (float, np.floating))

    def test_slew_ra_dec_ra_in_range(self, slew_setup):
        ra, dec = slew_setup.slew_ra_dec(1700000000.0)
        assert 0 <= ra < 360

    def test_slew_ra_dec_dec_in_range(self, slew_setup):
        ra, dec = slew_setup.slew_ra_dec(1700000000.0)
        assert -90 <= dec <= 90

    def test_slew_ra_dec_no_path_returns_ra_start(self, slew):
        slew.startra = 45.0
        slew.startdec = 30.0
        slew.slewstart = 1700000000.0
        ra, dec = slew.slew_ra_dec(1700000000.0)
        assert ra == 45.0

    def test_slew_ra_dec_no_path_returns_dec_start(self, slew):
        slew.startra = 45.0
        slew.startdec = 30.0
        slew.slewstart = 1700000000.0
        ra, dec = slew.slew_ra_dec(1700000000.0)
        assert dec == 30.0

    def test_slew_ra_dec_interpolation_mid_slew(self, slew_interpolation):
        ra, dec = slew_interpolation.slew_ra_dec(1700000050.0)
        assert isinstance(ra, (float, np.floating))
        assert isinstance(dec, (float, np.floating))
        assert 45.0 <= ra <= 90.0
        assert 30.0 <= dec <= 60.0

    def test_slew_ra_dec_modern_path_small_n(self, slew_modern_path):
        ra, dec = slew_modern_path.slew_ra_dec(1700000050.0)
        assert ra == 67.5
        assert dec == 45.0

    def test_slew_ra_dec_acs_returns_ra_float(self, slew_acs):
        ra, dec = slew_acs.slew_ra_dec(1700000050.0)
        assert isinstance(ra, (float, np.floating))

    def test_slew_ra_dec_acs_returns_dec_float(self, slew_acs):
        ra, dec = slew_acs.slew_ra_dec(1700000050.0)
        assert isinstance(dec, (float, np.floating))

    def test_slew_ra_dec_acs_ra_in_range(self, slew_acs):
        ra, dec = slew_acs.slew_ra_dec(1700000050.0)
        assert 0 <= ra < 360

    def test_slew_ra_dec_acs_dec_in_range(self, slew_acs):
        ra, dec = slew_acs.slew_ra_dec(1700000050.0)
        assert -90 <= dec <= 90

    def test_slew_ra_dec_interpolates_ra_at_start(self, slew_interp_start):
        ra, dec = slew_interp_start.slew_ra_dec(1700000000.0)
        assert np.isclose(ra, 0.0)

    def test_slew_ra_dec_interpolates_dec_at_start(self, slew_interp_start):
        ra, dec = slew_interp_start.slew_ra_dec(1700000000.0)
        assert np.isclose(dec, 0.0)


class TestCalcSlewtime:
    """Test calc_slewtime method."""

    def test_calc_slewtime_returns_slewtime(self, slew_calc_setup):
        slewtime = slew_calc_setup.calc_slewtime()
        assert slewtime == 50.0

    def test_calc_slewtime_sets_slewend(self, slew_calc_setup):
        slew_calc_setup.calc_slewtime()
        assert slew_calc_setup.slewend == 1700000050.0

    def test_calc_slewtime_sets_slewend_correctly(self, slew_calc_setup_alt):
        slew_calc_setup_alt.calc_slewtime()
        assert slew_calc_setup_alt.slewend == 1700000030.0

    @pytest.mark.parametrize("distance", [np.nan, -5.0])
    def test_calc_slewtime_handles_invalid_distance(self, slew, acs_config, distance):
        acs_config.predict_slew = Mock(
            return_value=(distance, (np.array([0.0]), np.array([0.0])))
        )
        acs_config.slew_time = Mock(return_value=0.0)
        slew.startra = 0.0
        slew.startdec = 0.0
        slew.endra = 10.0
        slew.enddec = 10.0
        slew.slewstart = 1700000000.0
        with pytest.raises(ValueError, match="Invalid slew distance"):
            slew.calc_slewtime()


class TestPredictSlew:
    """Test predict_slew method."""

    def test_predict_slew_calls_acs_predict_slew(self, slew_predict_setup):
        slew, ra_path, dec_path = slew_predict_setup
        slew.predict_slew()
        slew.acs_config.predict_slew.assert_called_once_with(
            45.0, 30.0, 90.0, 60.0, steps=20
        )

    def test_predict_slew_sets_slewdist(self, slew_predict_setup):
        slew, ra_path, dec_path = slew_predict_setup
        slew.predict_slew()
        assert slew.slewdist == 14.142

    def test_predict_slew_sets_path_ra_length(self, slew_predict_setup):
        slew, ra_path, dec_path = slew_predict_setup
        slew.predict_slew()
        assert len(slew.slewpath[0]) == 20

    def test_predict_slew_sets_path_dec_length(self, slew_predict_setup):
        slew, ra_path, dec_path = slew_predict_setup
        slew.predict_slew()
        assert len(slew.slewpath[1]) == 20

    def test_predict_slew_sets_path_ra_values(self, slew_predict_setup):
        slew, ra_path, dec_path = slew_predict_setup
        slew.predict_slew()
        assert np.allclose(slew.slewpath[0], ra_path)

    def test_predict_slew_sets_path_dec_values(self, slew_predict_setup):
        slew, ra_path, dec_path = slew_predict_setup
        slew.predict_slew()
        assert np.allclose(slew.slewpath[1], dec_path)



"""Unit tests for the DumbQueueScheduler class."""

from datetime import datetime, timedelta, timezone
from unittest.mock import Mock

import pytest  # type: ignore[import-untyped]
from astropy.time import Time  # type: ignore[import-untyped]

from conops import DAY_SECONDS, DumbQueueScheduler, Plan


class TestDumbQueueSchedulerInit:
    """Test DumbQueueScheduler initialization."""

    def test_init_default_queue_not_none(self, mock_queue):
        scheduler = DumbQueueScheduler(queue=mock_queue)
        assert scheduler.queue is not None

    def test_init_default_plan_not_none(self, mock_queue):
        scheduler = DumbQueueScheduler(queue=mock_queue)
        assert scheduler.plan is not None

    def test_init_default_begin_value(self, mock_queue):
        scheduler = DumbQueueScheduler(queue=mock_queue)
        assert scheduler.begin is None

    def test_init_default_end_value(self, mock_queue):
        scheduler = DumbQueueScheduler(queue=mock_queue)
        assert scheduler.end is None

    def test_init_with_custom_parameter_queue(
        self, mock_queue, scheduler_2022_100_len2
    ):
        assert scheduler_2022_100_len2.queue is mock_queue

    def test_init_with_custom_parameter_plan(self, mock_queue):
        plan = Plan()
        scheduler = DumbQueueScheduler(
            queue=mock_queue,
            plan=plan,
            begin=datetime(2022, 4, 10),
            end=datetime(2022, 4, 11),
        )
        assert scheduler.plan is plan

    def test_init_with_custom_parameter_begin(self, mock_queue):
        plan = Plan()
        begin = datetime(2022, 4, 10, tzinfo=timezone.utc)
        end = begin + timedelta(days=2)
        scheduler = DumbQueueScheduler(
            queue=mock_queue, plan=plan, begin=begin, end=end
        )
        assert scheduler.begin == begin

    def test_init_with_custom_parameter_end(self, mock_queue):
        plan = Plan()
        begin = datetime(2022, 4, 10, tzinfo=timezone.utc)
        end = begin + timedelta(days=2)
        scheduler = DumbQueueScheduler(
            queue=mock_queue, plan=plan, begin=begin, end=end
        )
        assert scheduler.end == end

    def test_init_creates_empty_plan(self):
        scheduler = DumbQueueScheduler(queue=Mock())
        assert len(scheduler.plan) == 0

    def test_init_creates_empty_queue_not_none(self):
        scheduler = DumbQueueScheduler(queue=Mock())
        assert scheduler.queue is not None


class TestDumbQueueSchedulerSchedule:
    """Test the schedule method."""

    def test_schedule_returns_plan(self, scheduler):
        result = scheduler.schedule()
        assert isinstance(result, Plan)

    def test_schedule_precondition_plan_nonempty(self, scheduler):
        scheduler.plan.extend([Mock()])  # Add dummy entry
        assert len(scheduler.plan) > 0

    def test_schedule_returns_plan_when_prepopulated(self, scheduler):
        scheduler.plan.extend([Mock()])  # Add dummy entry
        result = scheduler.schedule()
        assert isinstance(result, Plan)

    def test_schedule_with_single_target_returns_at_least_one_entry(
        self, scheduler, make_target, queue_get_from_list
    ):
        target = make_target(targetid=1, ra=45.0, dec=30.0, merit=100, ss_min=300)
        target.begin = int(scheduler.ustart)
        target.end = int(scheduler.ustart + 600)  # 10 minutes

        queue_get_from_list(scheduler, [target])

        result = scheduler.schedule()
        assert len(result) >= 1

    def test_schedule_with_multiple_targets_returns_at_least_one_entry(
        self, scheduler, make_targets, queue_get_from_list
    ):
        targets = make_targets(count=3)
        targets[0].begin = scheduler.ustart
        targets[0].end = scheduler.ustart + 600

        targets[1].begin = scheduler.ustart + 600
        targets[1].end = scheduler.ustart + 1200

        targets[2].begin = scheduler.ustart + 1200
        targets[2].end = scheduler.ustart + 1800

        queue_get_from_list(scheduler, targets)

        result = scheduler.schedule()
        assert len(result) >= 1

    def test_schedule_stops_when_queue_empty(self, scheduler):
        scheduler.queue.get = Mock(return_value=None)

        scheduler.schedule()
        scheduler.queue.get.assert_called()

    def test_schedule_respects_time_window(
        self, scheduler, make_target, queue_get_from_list
    ):
        target = make_target(targetid=1, ra=45.0, dec=30.0, merit=100, ss_min=300)

        queue_get_from_list(scheduler, [target])
        target.begin = scheduler.ustart
        target.end = scheduler.ustart + DAY_SECONDS * 2  # Beyond window

        result = scheduler.schedule()
        assert isinstance(result, Plan)


class TestDumbQueueSchedulerStartTime:
    """Test start time calculation."""

    def test_ustart_calculation(self, scheduler):
        scheduler.schedule()
        assert scheduler.ustart > 0

    @pytest.mark.parametrize(
        "begin_date",
        [
            datetime(2020, 1, 1, tzinfo=timezone.utc),
            datetime(2021, 4, 10, tzinfo=timezone.utc),
            datetime(2022, 12, 31, tzinfo=timezone.utc),
        ],
    )
    def test_different_begin_dates_set_ustart(self, begin_date):
        end = begin_date + timedelta(days=1)
        scheduler = DumbQueueScheduler(queue=Mock(), begin=begin_date, end=end)
        scheduler.queue.get = Mock(return_value=None)
        scheduler.schedule()
        assert scheduler.ustart > 0

    def test_multi_day_scheduling_end_preserved(self):
        begin = datetime(2021, 1, 4, tzinfo=timezone.utc)
        end = begin + timedelta(days=3)
        scheduler = DumbQueueScheduler(queue=Mock(), begin=begin, end=end)
        scheduler.queue.get = Mock(return_value=None)
        scheduler.schedule()
        assert scheduler.end == end


class TestDumbQueueSchedulerTargetProcessing:
    """Test target processing during scheduling."""

    def test_target_marked_done_after_scheduling(
        self, scheduler, make_target, queue_get_from_list
    ):
        target = make_target(targetid=1, ra=45.0, dec=30.0, merit=100)
        target.begin = scheduler.ustart
        target.end = scheduler.ustart + 600
        target.done = False

        queue_get_from_list(scheduler, [target])

        _ = scheduler.schedule()
        assert target.done is True

    def test_target_position_updated_during_scheduling_adds_entries(
        self, scheduler, make_targets, queue_get_from_list
    ):
        targets = make_targets(count=2)

        targets[0].begin = scheduler.ustart
        targets[0].end = scheduler.ustart + 600

        targets[1].begin = scheduler.ustart + 600
        targets[1].end = scheduler.ustart + 1200

        queue_get_from_list(scheduler, targets)
        result = scheduler.schedule()

        assert len(result) >= 1


class TestDumbQueueSchedulerEdgeCases:
    """Test edge cases and error handling."""

    def test_zero_duration_target_returns_plan(
        self, scheduler, make_target, queue_get_from_list
    ):
        target = make_target(targetid=1, ra=45.0, dec=30.0, merit=100)
        target.begin = scheduler.ustart
        target.end = scheduler.ustart  # Zero duration

        queue_get_from_list(scheduler, [target])

        result = scheduler.schedule()
        assert isinstance(result, Plan)

    def test_negative_duration_target_returns_plan(
        self, scheduler, make_target, queue_get_from_list
    ):
        target = make_target(targetid=1, ra=45.0, dec=30.0, merit=100)
        target.begin = scheduler.ustart
        target.end = scheduler.ustart - 100  # Negative duration

        queue_get_from_list(scheduler, [target])

        result = scheduler.schedule()
        assert isinstance(result, Plan)

    def test_very_long_scheduling_window(
        self, scheduler, make_target, queue_get_from_list
    ):
        # Extend the scheduling window to a full year
        scheduler.end = scheduler.begin + timedelta(days=365)

        target = make_target(targetid=1, ra=45.0, dec=30.0, merit=100)
        target.begin = scheduler.ustart
        target.end = scheduler.ustart + 600

        queue_get_from_list(scheduler, [target])
        result = scheduler.schedule()
        assert isinstance(result, Plan)

    def test_many_targets_in_queue_returns_plan(
        self, scheduler, make_targets, queue_get_from_list
    ):
        targets = make_targets(count=50)

        for i, target in enumerate(targets):
            target.begin = scheduler.ustart + i * 1000
            target.end = scheduler.ustart + (i + 1) * 1000

        queue_get_from_list(scheduler, targets)
        result = scheduler.schedule()
        assert isinstance(result, Plan)


class TestDumbQueueSchedulerIntegration:
    """Integration tests."""

    def test_full_scheduling_workflow_returns_plan(
        self, scheduler, make_targets, queue_get_from_list
    ):
        scheduler.ustart = Time("2021-01-04 00:00:00", scale="utc").unix

        targets = make_targets(count=2)

        for i, target in enumerate(targets):
            target.begin = scheduler.ustart + i * 1000
            target.end = scheduler.ustart + (i + 1) * 1000

        queue_get_from_list(scheduler, targets)
        result = scheduler.schedule()
        assert isinstance(result, Plan)

    def test_full_scheduling_workflow_all_entries_begin_after_ustart(
        self, scheduler, make_targets, queue_get_from_list
    ):
        scheduler.ustart = Time("2021-01-04 00:00:00", scale="utc").unix

        targets = make_targets(count=2)

        for i, target in enumerate(targets):
            target.begin = scheduler.ustart + i * 1000
            target.end = scheduler.ustart + (i + 1) * 1000

        queue_get_from_list(scheduler, targets)
        result = scheduler.schedule()
        assert all(entry.begin >= scheduler.ustart for entry in result.entries)

    def test_full_scheduling_workflow_all_entries_end_within_window(
        self, scheduler, make_targets, queue_get_from_list
    ):
        scheduler.ustart = Time("2021-01-04 00:00:00", scale="utc").unix

        targets = make_targets(count=2)

        for i, target in enumerate(targets):
            target.begin = scheduler.ustart + i * 1000
            target.end = scheduler.ustart + (i + 1) * 1000

        queue_get_from_list(scheduler, targets)
        result = scheduler.schedule()
        window_end = scheduler.end.timestamp()
        assert all(entry.end <= window_end for entry in result.entries)

    def test_scheduling_with_position_tracking_records_positions(
        self, scheduler, make_targets, queue_get_from_list
    ):
        targets = make_targets(count=3)

        for i, target in enumerate(targets):
            target.begin = scheduler.ustart + i * 1000
            target.end = scheduler.ustart + (i + 1) * 1000

        positions = queue_get_from_list(scheduler, targets, track_positions=True)
        _ = scheduler.schedule()
        assert len(positions) >= 1

    def test_scheduling_with_position_tracking_initial_position(
        self, scheduler, make_targets, queue_get_from_list
    ):
        targets = make_targets(count=3)

        for i, target in enumerate(targets):
            target.begin = scheduler.ustart + i * 1000
            target.end = scheduler.ustart + (i + 1) * 1000

        positions = queue_get_from_list(scheduler, targets, track_positions=True)
        _ = scheduler.schedule()
        assert positions[0] == (0.0, 0.0)

    def test_plan_entries_in_sequence_type(
        self, scheduler, make_targets, queue_get_from_list
    ):
        targets = make_targets(count=3)

        base_time = scheduler.ustart
        for i, target in enumerate(targets):
            target.begin = base_time + i * 1000
            target.end = base_time + (i + 1) * 1000

        queue_get_from_list(scheduler, targets)
        plan = scheduler.schedule()

        assert isinstance(plan, Plan)

    def test_plan_entries_in_time_sequence(
        self, scheduler, make_targets, queue_get_from_list
    ):
        targets = make_targets(count=3)

        base_time = scheduler.ustart
        for i, target in enumerate(targets):
            target.begin = base_time + i * 1000
            target.end = base_time + (i + 1) * 1000

        queue_get_from_list(scheduler, targets)
        plan = scheduler.schedule()

        assert all(
            plan.entries[i].end <= plan.entries[i + 1].begin
            for i in range(max(0, len(plan.entries) - 1))
        )


class TestDumbQueueSchedulerStateManagement:
    """Test state management and plan reuse."""

    def test_plan_reset_between_runs_independent_plans(
        self, scheduler, make_target, queue_get_from_list
    ):
        target = make_target(targetid=1, ra=45.0, dec=30.0, merit=100)
        target.begin = scheduler.ustart
        target.end = scheduler.ustart + 600

        queue_get_from_list(scheduler, [target])
        plan1 = scheduler.schedule()

        # Ensure a fresh plan will be returned on a new run
        # Reuse the queue behavior by re-attaching the same target
        plan2_targets = [target]
        queue_get_from_list(scheduler, plan2_targets)
        plan2 = scheduler.schedule()

        assert plan1 is not plan2

    def test_scheduler_parameters_preserved_begin(self, scheduler):
        original_begin = scheduler.begin
        scheduler.queue.get = Mock(return_value=None)
        scheduler.schedule()
        assert scheduler.begin == original_begin

    def test_scheduler_parameters_preserved_end(self, scheduler):
        original_end = scheduler.end
        scheduler.queue.get = Mock(return_value=None)
        scheduler.schedule()
        assert scheduler.end == original_end


class TestDumbQueueSchedulerConfiguration:
    """Test configuration options."""

    @pytest.mark.parametrize("begin_year", [2000, 2021, 2050])
    def test_begin_parameter(self, begin_year):
        begin = datetime(begin_year, 1, 1, tzinfo=timezone.utc)
        end = begin + timedelta(days=1)
        scheduler = DumbQueueScheduler(queue=Mock(), begin=begin, end=end)
        assert scheduler.begin == begin

    @pytest.mark.parametrize("end_year", [2000, 2021, 2050])
    def test_end_parameter(self, end_year):
        begin = datetime(2021, 1, 1, tzinfo=timezone.utc)
        end = datetime(end_year, 1, 2, tzinfo=timezone.utc)
        scheduler = DumbQueueScheduler(queue=Mock(), begin=begin, end=end)
        assert scheduler.end == end



"""Communications system configuration for spacecraft downlink and uplink."""

from typing import Any, Literal

from pydantic import BaseModel, Field, model_validator

from ..common.enums import AntennaType, Polarization


class BandCapability(BaseModel):
    """Defines a communication band's data rate capabilities."""

    band: Literal["S", "X", "Ka", "Ku", "L", "C", "K"]  # Common spacecraft bands
    uplink_rate_mbps: float = Field(
        default=0.0, description="Uplink data rate in Mbps", ge=0.0
    )
    downlink_rate_mbps: float = Field(
        default=0.0, description="Downlink data rate in Mbps", ge=0.0
    )

    @model_validator(mode="before")
    @classmethod
    def _apply_standard_band_defaults(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Fill in standard per-band defaults when rates are not provided.

        Only applies when `uplink_rate_mbps` and/or `downlink_rate_mbps` are
        not passed in the input data. Custom-provided values are preserved.
        """
        if isinstance(data, dict):
            band: str | None = data.get("band")
            defaults = {
                "S": {"uplink": 2.0, "downlink": 10.0},
                "X": {"uplink": 10.0, "downlink": 150.0},
                "Ka": {"uplink": 20.0, "downlink": 300.0},
                "Ku": {"uplink": 5.0, "downlink": 50.0},
                "L": {"uplink": 0.5, "downlink": 1.0},
                "C": {"uplink": 2.0, "downlink": 20.0},
                "K": {"uplink": 15.0, "downlink": 200.0},
            }
            if band in defaults:
                if (
                    "uplink_rate_mbps" not in data
                    or data.get("uplink_rate_mbps") is None
                ):
                    data["uplink_rate_mbps"] = defaults[band]["uplink"]
                if (
                    "downlink_rate_mbps" not in data
                    or data.get("downlink_rate_mbps") is None
                ):
                    data["downlink_rate_mbps"] = defaults[band]["downlink"]
        return data


class AntennaPointing(BaseModel):
    """Defines antenna pointing configuration based on antenna type."""

    antenna_type: AntennaType = Field(
        default=AntennaType.FIXED,
        description="Type of antenna: omni, fixed, or gimbaled",
    )

    # Fixed antenna pointing (body-fixed direction)
    fixed_azimuth_deg: float = Field(
        default=0.0,
        description="Azimuth angle in spacecraft body frame (deg). 0=nadir, 180=zenith",
        ge=0.0,
        le=360.0,
    )
    fixed_elevation_deg: float = Field(
        default=0.0,
        description="Elevation angle in spacecraft body frame (deg)",
        ge=-90.0,
        le=90.0,
    )

    # Gimbaled antenna range (angular range of motion from boresight)
    gimbal_range_deg: float = Field(
        default=0.0,
        description="Angular range for gimbaled antenna (deg) from boresight",
        ge=0.0,
        le=180.0,
    )

    def is_nadir_pointing(self) -> bool:
        """Check if antenna is nadir-pointing (opposite of telescope)."""
        return (
            self.antenna_type == AntennaType.FIXED
            and self.fixed_azimuth_deg == 0.0
            and self.fixed_elevation_deg == 0.0
        )


class CommunicationsSystem(BaseModel):
    """Onboard communications system configuration.

    Defines band capabilities, data rates, antenna type, pointing,
    and signal requirements for ground station passes.
    """

    name: str = Field(default="Default Comms", description="Communications system name")

    # Band capabilities (multiple bands supported)
    band_capabilities: list[BandCapability] = Field(
        default_factory=lambda: [
            BandCapability(band="S", uplink_rate_mbps=2.0, downlink_rate_mbps=10.0)
        ],
        description="List of supported frequency bands with data rates",
    )

    # Antenna configuration
    antenna_pointing: AntennaPointing = Field(
        default_factory=lambda: AntennaPointing(antenna_type=AntennaType.FIXED),
        description="Antenna pointing configuration",
    )

    # Signal quality requirement
    pointing_accuracy_deg: float = Field(
        default=5.0,
        description="Maximum pointing error (deg) to maintain good signal",
        ge=0.0,
        le=180.0,
    )

    # Antenna polarization
    polarization: Polarization = Field(
        default=Polarization.CIRCULAR_RIGHT,
        description="Antenna polarization type",
    )

    def get_band(self, band: str) -> BandCapability | None:
        """Get capability for a specific band.

        Args:
            band: Band identifier (e.g., "S", "X", "Ka")

        Returns:
            BandCapability if found, None otherwise
        """
        for capability in self.band_capabilities:
            if capability.band == band:
                return capability
        return None

    def get_downlink_rate(self, band: str) -> float:
        """Get downlink rate for a specific band in Mbps.

        Args:
            band: Band identifier (e.g., "S", "X", "Ka")

        Returns:
            Downlink rate in Mbps, or 0.0 if band not supported
        """
        capability = self.get_band(band)
        return capability.downlink_rate_mbps if capability else 0.0

    def get_uplink_rate(self, band: str) -> float:
        """Get uplink rate for a specific band in Mbps.

        Args:
            band: Band identifier (e.g., "S", "X", "Ka")

        Returns:
            Uplink rate in Mbps, or 0.0 if band not supported
        """
        capability = self.get_band(band)
        return capability.uplink_rate_mbps if capability else 0.0

    def can_communicate(self, pointing_error_deg: float) -> bool:
        """Check if communication is possible given pointing error.

        Args:
            pointing_error_deg: Pointing error in degrees

        Returns:
            True if pointing error is within acceptable range
        """
        # Omni antenna can always communicate
        if self.antenna_pointing.antenna_type == AntennaType.OMNI:
            return True

        # For fixed/gimbaled, check against pointing accuracy requirement
        return pointing_error_deg <= self.pointing_accuracy_deg



"""DITL event logging system."""

from pydantic import BaseModel, Field

from ..common import ACSMode
from .ditl_event import DITLEvent
from .ditl_log_store import DITLLogStore


class DITLLog(BaseModel):
    """
    Container for DITL simulation events with logging and printing methods.

    Attributes:
        events: List of DITLEvent objects logged during simulation
        run_id: Optional identifier for this DITL run; used by stores
        store: Optional persistent store to which events are appended
    """

    events: list[DITLEvent] = Field(
        default_factory=list, description="List of logged events"
    )
    run_id: str | None = Field(default=None, description="Unique ID for this DITL run")
    store: DITLLogStore | None = Field(
        default=None, description="Optional persistent store"
    )

    def log_event(
        self,
        utime: float,
        event_type: str,
        description: str,
        obsid: int | None = None,
        acs_mode: ACSMode | None = None,
    ) -> None:
        """
        Log a DITL event.

        Parameters
        ----------
        utime : float
            Unix timestamp of the event
        event_type : str
            Category of event (e.g., 'PASS', 'SLEW', 'OBSERVATION', 'ERROR', 'INFO')
        description : str
            Human-readable description of the event
        obsid : int | None
            Optional observation ID associated with the event
        acs_mode : ACSMode | None
            Optional ACS mode at the time of the event
        """
        event = DITLEvent.from_utime(
            utime=utime,
            event_type=event_type,
            description=description,
            obsid=obsid,
            acs_mode=acs_mode,
        )
        self.events.append(event)
        # If a store is configured with a run_id, persist as we go
        if self.store is not None and self.run_id is not None:
            try:
                self.store.add_event(self.run_id, event)
            except Exception:
                # Non-fatal: keep in-memory log even if persistence fails
                pass

    def print_log(self) -> None:
        """Print the DITL event log to stdout."""
        for event in self.events:
            print(str(event))

    def clear(self) -> None:
        """Clear all logged events."""
        self.events.clear()

    def __len__(self) -> int:
        """Return the number of logged events."""
        return len(self.events)

    def __getitem__(self, index: int) -> DITLEvent:
        """Get event by index."""
        return self.events[index]

    def flush_to_store(self) -> None:
        """Persist all current events to the configured store, if any.

        Safe to call multiple times; duplicates will accumulate if called
        repeatedly without store-level deduplication.
        """
        if self.store is not None and self.run_id is not None and self.events:
            try:
                self.store.add_events(self.run_id, self.events)
            except Exception:
                pass



"""Tests for conops.common module."""

from conops import (
    ACSMode,
    givename,
    ics_date_conv,
    unixtime2date,
    unixtime2yearday,
)


class TestACSMode:
    """Test ACS Mode enum."""

    def test_acs_mode_science(self):
        """Test SCIENCE mode value."""
        assert ACSMode.SCIENCE == 0

    def test_acs_mode_slewing(self):
        """Test SLEWING mode value."""
        assert ACSMode.SLEWING == 1

    def test_acs_mode_saa(self):
        """Test SAA mode value."""
        assert ACSMode.SAA == 2

    def test_acs_mode_pass(self):
        """Test PASS mode value."""
        assert ACSMode.PASS == 3

    def test_acs_mode_charging(self):
        """Test CHARGING mode value."""
        assert ACSMode.CHARGING == 4

    def test_acs_mode_is_int(self):
        """Test that ACS mode values are integers."""
        assert isinstance(ACSMode.SCIENCE, int)


class TestGivename:
    """Test givename coordinate naming function."""

    def test_givename_positive_dec(self):
        """Test givename with positive declination."""
        # RA = 45 degrees = 3 hours, Dec = 30 degrees
        name = givename(45.0, 30.0)
        assert "J" in name
        assert "03" in name  # 45/15 = 3 hours
        assert "30" in name  # Dec degrees

    def test_givename_negative_dec(self):
        """Test givename with negative declination."""
        # RA = 60 degrees = 4 hours, Dec = -45 degrees
        name = givename(60.0, -45.0)
        assert "J" in name
        assert "-" in name  # Negative indicator

    def test_givename_with_stem(self):
        """Test givename with stem prefix."""
        name = givename(45.0, 30.0, stem="TEST")
        assert "TEST" in name

    def test_givename_zero_ra_dec(self):
        """Test givename with zero RA and Dec."""
        name = givename(0.0, 0.0)
        assert isinstance(name, str)
        assert len(name) > 0


class TestUnixtime2date:
    """Test Unix timestamp to date conversion."""

    def test_unixtime2date_known_time(self):
        """Test conversion of known Unix timestamp."""
        # Unix timestamp for 2023-01-01 00:00:00 UTC
        utime = 1672531200.0
        date_str = unixtime2date(utime)
        assert isinstance(date_str, str)
        assert "2023" in date_str

    def test_unixtime2date_format(self):
        """Test that date format is correct."""
        utime = 1700000000.0
        date_str = unixtime2date(utime)
        # Format should be YYYY-DDD-HH:MM:SS
        parts = date_str.split("-")
        assert len(parts) >= 2
        assert len(parts[0]) == 4  # Year is 4 digits


class TestIcsDateConv:
    """Test ICS date conversion function."""

    def test_ics_date_conv_valid_date(self):
        """Test ICS date conversion with valid date."""
        # Format: "YYYY/DDD-HH:MM:SS"
        ics_date = "2023/001-00:00:00"
        unix_time = ics_date_conv(ics_date)
        assert isinstance(unix_time, (int, float))

    def test_ics_date_conv_another_date(self):
        """Test ICS date conversion with another date."""
        ics_date = "2023/100-12:30:45"
        unix_time = ics_date_conv(ics_date)
        assert isinstance(unix_time, (int, float))
        assert unix_time > 0


class TestUnixtimeToYearday:
    """Test Unix timestamp to year and day of year conversion."""

    def test_unixtime2yearday_known_time(self):
        """Test conversion of known Unix timestamp."""
        # Unix timestamp for 2023-01-01 00:00:00 UTC
        utime = 1672531200.0
        year, day = unixtime2yearday(utime)
        assert year == 2023
        assert day == 1

    def test_unixtime2yearday_mid_year(self):
        """Test conversion for mid-year date."""
        # Unix timestamp for 2023-07-02 (approximately day 183)
        utime = 1688169600.0
        year, day = unixtime2yearday(utime)
        assert year == 2023
        assert day > 100  # Mid-year



from collections import Counter
from datetime import datetime

import numpy as np

from ..config.config import MissionConfig


class DITLStats:
    """Mixin class providing statistics printing functionality for DITL simulations."""

    config: MissionConfig
    begin: datetime
    end: datetime
    step_size: int
    utime: list[float]
    mode: list[int]
    ra: list[float]
    dec: list[float]
    roll: list[float]
    obsid: list[int]
    panel: list[float]
    power: list[float]
    batterylevel: list[float]
    recorder_fill_fraction: list[float]
    data_generated_gb: list[float]
    data_downlinked_gb: list[float]

    def print_statistics(self) -> None:
        """Print comprehensive statistics about the DITL simulation.

        Displays information about:
        - Simulation time period and duration
        - Mode distribution (time spent in each ACS mode)
        - Observation statistics (unique targets, total observations)
        - Power and battery statistics
        - Solar panel performance
        - Queue information (if available)
        - ACS commands (if available)
        - Ground station pass statistics (if available)
        """
        from ..common import ACSMode

        # Basic simulation info
        print("=" * 70)
        print("DITL SIMULATION STATISTICS")
        print("=" * 70)
        print(f"\nConfiguration: {self.config.name}")
        print(f"Start Time: {self.begin.strftime('%Y-%m-%d %H:%M:%S UTC')}")
        print(f"End Time: {self.end.strftime('%Y-%m-%d %H:%M:%S UTC')}")
        duration_hours = (self.end - self.begin).total_seconds() / 3600
        print(f"Duration: {duration_hours:.2f} hours ({duration_hours / 24:.2f} days)")
        print(f"Time Steps: {len(self.utime)}")
        print(f"Step Size: {self.step_size} seconds")

        # Mode statistics
        print("\n" + "-" * 70)
        print("MODE DISTRIBUTION")
        print("-" * 70)
        if self.mode:
            mode_counts = Counter(self.mode)
            total_steps = len(self.mode)
            print(f"{'Mode':<20} {'Count':<10} {'Percentage':<12} {'Time (hours)':<15}")
            print("-" * 70)
            for mode_val, count in sorted(mode_counts.items()):
                mode_name = (
                    ACSMode(mode_val).name
                    if mode_val in [m.value for m in ACSMode]
                    else f"UNKNOWN({mode_val})"
                )
                percentage = (count / total_steps) * 100
                time_hours = (count * self.step_size) / 3600
                print(
                    f"{mode_name:<20} {count:<10} {percentage:>6.2f}%      {time_hours:>10.2f}"
                )

        # Observation statistics
        print("\n" + "-" * 70)
        print("OBSERVATION STATISTICS")
        print("-" * 70)
        if self.obsid:
            unique_obsids = set(self.obsid)
            # Filter out special ObsIDs (like 0 or 999xxx for charging)
            science_obsids = [o for o in unique_obsids if o > 0 and o < 999000]
            print(f"Total Unique Observations: {len(science_obsids)}")
            print(
                f"Total Observation Steps: {sum(1 for o in self.obsid if 0 < o < 999000)}"
            )

            if science_obsids:
                # Count time per obsid
                obsid_counts = Counter([o for o in self.obsid if 0 < o < 999000])
                print("\nTop 10 Observations by Time:")
                print(f"{'ObsID':<10} {'Steps':<10} {'Time (hours)':<15}")
                print("-" * 35)
                for obsid, count in obsid_counts.most_common(10):
                    time_hours = (count * self.step_size) / 3600
                    print(f"{obsid:<10} {count:<10} {time_hours:>10.2f}")

        # Pointing statistics
        print("\n" + "-" * 70)
        print("POINTING STATISTICS")
        print("-" * 70)
        if self.ra and self.dec:
            print(f"Total Pointing Updates: {len(self.ra)}")
            print(f"RA Range: {min(self.ra):.2f}° to {max(self.ra):.2f}°")
            print(f"Dec Range: {min(self.dec):.2f}° to {max(self.dec):.2f}°")
            if hasattr(self, "roll") and self.roll:
                print(f"Roll Range: {min(self.roll):.2f}° to {max(self.roll):.2f}°")

        # Battery statistics
        print("\n" + "-" * 70)
        print("POWER AND BATTERY STATISTICS")
        print("-" * 70)
        if self.batterylevel:
            battery_capacity = getattr(
                self.config.battery,
                "watthour",
                getattr(self.config.battery, "capacity", None),
            )
            if battery_capacity is not None:
                print(f"Battery Capacity: {battery_capacity:.2f} Wh")
            print(f"Initial Charge: {self.batterylevel[0] * 100:.1f}%")
            print(f"Final Charge: {self.batterylevel[-1] * 100:.1f}%")
            print(f"Min Charge: {min(self.batterylevel) * 100:.1f}%")
            print(f"Max Charge: {max(self.batterylevel) * 100:.1f}%")
            print(f"Avg Charge: {np.mean(self.batterylevel) * 100:.1f}%")
            max_dod = self.config.battery.max_depth_of_discharge
            print(f"Max Depth of Discharge: {max_dod * 100:.1f}%")
            violations = sum(1 for bl in self.batterylevel if bl < max_dod)
            if violations > 0:
                print(
                    f"⚠️  DoD Violations: {violations} steps ({violations / len(self.batterylevel) * 100:.2f}%)"
                )

        # Charge state statistics
        if hasattr(self, "charge_state") and self.charge_state:
            from ..common import ChargeState

            print("\nBattery Charging State Distribution:")
            charge_state_counts = Counter(self.charge_state)
            total_steps = len(self.charge_state)
            print(
                f"{'State':<20} {'Count':<10} {'Percentage':<12} {'Time (hours)':<15}"
            )
            print("-" * 70)
            for state_val, count in sorted(charge_state_counts.items()):
                state_name = (
                    ChargeState(state_val).name
                    if state_val in [s.value for s in ChargeState]
                    else f"UNKNOWN({state_val})"
                )
                percentage = (count / total_steps) * 100
                time_hours = (count * self.step_size) / 3600
                print(
                    f"{state_name:<20} {count:<10} {percentage:>6.2f}%      {time_hours:>10.2f}"
                )

        if hasattr(self, "power") and self.power:
            print("\nPower Consumption:")
            print(f"  Average: {np.mean(self.power):.2f} W")
            print(f"  Peak: {max(self.power):.2f} W")
            print(f"  Minimum: {min(self.power):.2f} W")

            # Subsystem power breakdown if available
            if (
                hasattr(self, "power_bus")
                and hasattr(self, "power_payload")
                and self.power_bus
                and self.power_payload
            ):
                print("\n  Subsystem Breakdown:")
                avg_bus = np.mean(self.power_bus)
                avg_payload = np.mean(self.power_payload)
                print(
                    f"    Bus Average: {avg_bus:.2f} W ({avg_bus / np.mean(self.power) * 100:.1f}%)"
                )
                print(
                    f"    Payload Average: {avg_payload:.2f} W ({avg_payload / np.mean(self.power) * 100:.1f}%)"
                )
                print(f"    Bus Peak: {max(self.power_bus):.2f} W")
                print(f"    Payload Peak: {max(self.power_payload):.2f} W")

        if hasattr(self, "panel_power") and self.panel_power:
            print("\nSolar Panel Generation:")
            print(f"  Average: {np.mean(self.panel_power):.2f} W")
            print(f"  Peak: {max(self.panel_power):.2f} W")
            total_generated = sum(self.panel_power) * self.step_size / 3600  # Wh
            total_consumed = sum(self.power) * self.step_size / 3600  # Wh
            print(f"  Total Generated: {total_generated:.2f} Wh")
            print(f"  Total Consumed: {total_consumed:.2f} Wh")
            print(f"  Net Energy: {total_generated - total_consumed:.2f} Wh")

        if hasattr(self, "panel") and self.panel:
            print("\nSolar Panel Illumination:")
            avg_illumination = np.mean(self.panel) * 100
            print(f"  Average: {avg_illumination:.1f}%")
            eclipse_steps = sum(1 for p in self.panel if p < 0.01)
            print(
                f"  Eclipse Time: {eclipse_steps * self.step_size / 3600:.2f} hours ({eclipse_steps / len(self.panel) * 100:.1f}%)"
            )

        # Data Management statistics
        if (
            hasattr(self, "recorder_volume_gb")
            and self.recorder_volume_gb
            and self.config.recorder is not None
        ):
            print("\n" + "-" * 70)
            print("DATA MANAGEMENT STATISTICS")
            print("-" * 70)
            print(f"Recorder Capacity: {self.config.recorder.capacity_gb:.2f} Gb")
            print(f"Initial Volume: {self.recorder_volume_gb[0]:.2f} Gb")
            print(f"Final Volume: {self.recorder_volume_gb[-1]:.2f} Gb")
            print(f"Peak Volume: {max(self.recorder_volume_gb):.2f} Gb")

            if self.recorder_fill_fraction:
                print("\nFill Level:")
                print(f"  Initial: {self.recorder_fill_fraction[0] * 100:.1f}%")
                print(f"  Final: {self.recorder_fill_fraction[-1] * 100:.1f}%")
                print(f"  Peak: {max(self.recorder_fill_fraction) * 100:.1f}%")
                print(f"  Average: {np.mean(self.recorder_fill_fraction) * 100:.1f}%")

            if self.data_generated_gb:
                total_generated = (
                    self.data_generated_gb[-1] if self.data_generated_gb else 0
                )
                print(f"\nData Generated: {total_generated:.2f} Gb")

                # Calculate generation rate
                duration_hours = (self.end - self.begin).total_seconds() / 3600
                if duration_hours > 0:
                    avg_rate = total_generated / duration_hours
                    print(
                        f"  Average Rate: {avg_rate:.3f} Gb/hour ({avg_rate * 1000:.2f} Mbps)"
                    )

            if self.data_downlinked_gb:
                total_downlinked = (
                    self.data_downlinked_gb[-1] if self.data_downlinked_gb else 0
                )
                print(f"\nData Downlinked: {total_downlinked:.2f} Gb")

                # Calculate downlink efficiency
                if hasattr(self, "data_generated_gb") and self.data_generated_gb:
                    total_generated = self.data_generated_gb[-1]
                    if total_generated > 0:
                        efficiency = (total_downlinked / total_generated) * 100
                        print(f"  Downlink Efficiency: {efficiency:.1f}%")

                # Calculate downlink rate
                duration_hours = (self.end - self.begin).total_seconds() / 3600
                if duration_hours > 0:
                    avg_rate = total_downlinked / duration_hours
                    print(
                        f"  Average Rate: {avg_rate:.3f} Gb/hour ({avg_rate * 1000:.2f} Mbps)"
                    )

            # Recorder alert statistics
            if hasattr(self, "recorder_alert") and self.recorder_alert:
                alert_counts = Counter(self.recorder_alert)
                print("\nRecorder Alerts:")
                print(
                    f"  Yellow Threshold: {self.config.recorder.yellow_threshold * 100:.0f}%"
                )
                print(
                    f"  Red Threshold: {self.config.recorder.red_threshold * 100:.0f}%"
                )

                yellow_count = alert_counts.get(1, 0)  # 1 = yellow alert
                red_count = alert_counts.get(2, 0)  # 2 = red alert
                total_steps = len(self.recorder_alert)

                if yellow_count > 0:
                    yellow_time = yellow_count * self.step_size / 3600
                    print(
                        f"  Yellow Alerts: {yellow_count} steps ({yellow_time:.2f} hours, {yellow_count / total_steps * 100:.1f}%)"
                    )
                else:
                    print("  Yellow Alerts: None")

                if red_count > 0:
                    red_time = red_count * self.step_size / 3600
                    print(
                        f"  Red Alerts: {red_count} steps ({red_time:.2f} hours, {red_count / total_steps * 100:.1f}%)"
                    )
                else:
                    print("  Red Alerts: None")

        # Queue statistics (if available)
        if hasattr(self, "queue"):
            print("\n" + "-" * 70)
            print("TARGET QUEUE STATISTICS")
            print("-" * 70)
            print(f"Total Targets in Queue: {len(self.queue.targets)}")
            completed = sum(1 for t in self.queue.targets if getattr(t, "done", False))
            print(f"Completed Targets: {completed}")
            print(f"Remaining Targets: {len(self.queue.targets) - completed}")

        # ACS Command statistics (if available)
        if hasattr(self, "acs") and hasattr(self.acs, "commands"):
            print("\n" + "-" * 70)
            print("ACS COMMAND STATISTICS")
            print("-" * 70)
            cmd_counts = Counter([cmd.command_type for cmd in self.acs.commands])
            print(f"Total ACS Commands: {len(self.acs.commands)}")
            print(f"\n{'Command Type':<25} {'Count':<10}")
            print("-" * 35)
            for cmd_type, count in sorted(
                cmd_counts.items(), key=lambda x: x[1], reverse=True
            ):
                print(f"{cmd_type.name:<25} {count:<10}")

        # Ground station pass statistics (if available)
        if hasattr(self, "executed_passes") and len(self.executed_passes.passes) > 0:
            print("\n" + "-" * 70)
            print("GROUND STATION PASS STATISTICS")
            print("-" * 70)
            print(f"Total Passes Executed: {len(self.executed_passes.passes)}")
            total_pass_time = (
                sum((p.end - p.begin) for p in self.executed_passes.passes) / 3600
            )
            print(f"Total Pass Time: {total_pass_time:.2f} hours")

        print("\n" + "=" * 70)
